import unittest
from tests.test.RSC_PATHS import *
from tests.test.profiling import profile

from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj
from reamber.algorithms.analysis.describe.describe import describe

class TestDescribe(unittest.TestCase):

    # @profile
    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_ICFITU)
        describe(m, s=None)

    # @profile
    def test_qua(self):
        m = QuaMapObj()
        m.readFile(QUA_NEURO_CLOUD)
        describe(m, s=None)

    # @profile
    def test_sm(self):
        s = SMMapSetObj()
        s.readFile(SM_CARAVAN)
        describe(s.maps[0], s)

    # @profile
    def test_o2j(self):
        s = O2JMapSetObj()
        s.readFile(O2J_FLY_MAGPIE_OJN)
        describe(s.maps[2], s)

if __name__ == '__main__':
    unittest.main()



_MAPS_DIR = "../rsc/maps/"

_O2J_DIR = _MAPS_DIR + "o2jam/"
_OSU_DIR = _MAPS_DIR + "osu/"
_QUA_DIR = _MAPS_DIR + "qua/"
_SM_DIR = _MAPS_DIR + "sm/"

O2J_CHECK_IT_OUT_OJN = _O2J_DIR + "o2ma120.ojn"
O2J_CHECK_IT_OUT_OJM = _O2J_DIR + "o2ma120.ojm "
O2J_FLY_MAGPIE_OJN = _O2J_DIR + "o2ma178.ojn"
O2J_FLY_MAGPIE_OJM = _O2J_DIR + "o2ma178.ojm"

OSU_ADDICTION_CUT = _OSU_DIR + "AddictionCut.osu"  # Small file to test quickly
OSU_AVENGER_HITSOUNDABLE = _OSU_DIR + "AvengerHitsoundable.osu"
OSU_AVENGER_HITSOUNDFILE = _OSU_DIR + "AvengerHitsoundFile.osu"
OSU_CARAVAN = _OSU_DIR + "Caravan.osu"
OSU_CHECK_IT_OUT = _OSU_DIR + "CheckItOut.osu"
OSU_DOPPELGANGER = _OSU_DIR + "DoppelGanger.osu"
OSU_ESCAPES = _OSU_DIR + "Escapes.osu"
OSU_FOOL_MOON_NIGHT = _OSU_DIR + "FoolMoonNight.osu"
OSU_GRAVITY = _OSU_DIR + "Gravity.osu"
OSU_ICFITU = _OSU_DIR + "ICFITU.osu"
OSU_JOUNETSU = _OSU_DIR + "JounetsuTairiku.osu"
OSU_LAST_RESORT = _OSU_DIR + "LastResort.osu"
OSU_LEG = _OSU_DIR + "LEG.osu"
OSU_LN_DAN_14 = _OSU_DIR + "LNDan14.osu"
OSU_LN_DAN_14 = _OSU_DIR + "LNDan15.osu"
OSU_PLANET_SHAPER = _OSU_DIR + "PLANETSHAPER.osu"
OSU_REBIRTH = _OSU_DIR + "Rebirth.osu"
OSU_TRIBAL_TRIAL_EXH = _OSU_DIR + "TribalTrialEXH.osu"
OSU_TRIBAL_TRIAL_MX = _OSU_DIR + "TribalTrialMX.osu"
OSU_WORLD_FRAGMENTS = _OSU_DIR + "WorldFragments.osu"
OSU_ZENITHALIZE_19 = _OSU_DIR + "ZENITHALIZE_19.osu"

QUA_NEURO_CLOUD = _QUA_DIR + "NeuroCloud.qua"

SM_CARAVAN = _SM_DIR + "Caravan.sm"
SM_ESCAPES = _SM_DIR + "Escapes.sm"
SM_GRAVITY = _SM_DIR + "Gravity.sm"
SM_ICFITU = _SM_DIR + "ICFITU.sm"



from __future__ import annotations
from reamber.osu.lists.notes.OsuNoteList import OsuNoteList
from reamber.osu.OsuHitObj import OsuHitObj
from typing import List


class OsuHitList(List[OsuHitObj], OsuNoteList):

    def _upcast(self, objList: List = None) -> OsuHitList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: OsuHitList
        """
        return OsuHitList(objList)

    def data(self) -> List[OsuHitObj]:
        return self

    @staticmethod
    def readEditorString(s: str) -> OsuHitList:
        """ Reads an editor string, must be of the correct format.

        i.e. XX:XX:XXX(OFFSET|COL, OFFSET|COL, ...) -

        :param s: The editor string
        :return: Returns this class initialized
        """
        return OsuHitList([OsuHitObj(offset=float(note.split("|")[0]),
                                     column=int(note.split("|")[1]))
                           for note in s[s.find("(") + 1: s.find(")")].split(",")])



""" A Scroll Velocity Sequence is a 'property' class, where it holds information on how a SV should
be created.

More information and traits can be added onto this Class inside the parts package within this directory
"""

from __future__ import annotations

from reamber.osu.lists.OsuSvList import OsuSvList, OsuSvObj

from typing import List, overload, Tuple, Type
from reamber.algorithms.generate.sv.SvObj import SvObj
from reamber.algorithms.analysis.generic.activity import activity
from reamber.algorithms.generate.sv.SvIO import SvIO
from reamber.base.lists.TimedList import TimedList
from dataclasses import asdict
from copy import deepcopy
import logging

log = logging.getLogger(__name__)


class SvSequence(List[SvObj], TimedList, SvIO):

    def _upcast(self, objList: List = None):
        return SvSequence(objList)
    def data(self) -> List[SvObj]:
        return self
    def deepcopy(self) -> SvSequence:
        return deepcopy(self)

    @overload
    def __init__(self, list_: List[SvObj] = None): ...
    @overload
    def __init__(self, list_: List[Tuple[float, float, bool]] = None): ...
    @overload
    def __init__(self, list_: List[Tuple[float, float]] = None): ...
    @overload
    def __init__(self, list_: List[float] = None): ...
    def __init__(self, list_: List = None):
        """ Multiple ways to initialize the sequence

        1. `SvSequence([SvObj, SvObj, ...])`
        2. `SvSequence([(offset, sv), (offset, sv), ...])`
        3. `SvSequence([offset, offset, ...])`

        These also can be mixed.

        4. `SvSequence([SvObj, (offset, sv), offset, ...])`
        """

        if list_ is not None:
            for i in range(len(list_)):
                item = list_[i]
                if isinstance(item, Tuple):
                    if len(item) == 2:
                        list_[i] = SvObj(offset=item[0], multiplier=item[1])
                    elif len(item) == 3:
                        list_[i] = SvObj(offset=item[0], multiplier=item[1], fixed=item[2])
                elif isinstance(item, float) or isinstance(item, int):
                    list_[i] = SvObj(offset=item)
                elif not isinstance(item, SvObj):
                    raise TypeError(f"Unexpected Type in list. {item}")

            list.__init__(self, list_)
        else:
            list.__init__(self, [])

    def rescale(self, firstOffset: float, lastOffset: float, inplace: bool = False) -> SvSequence or None:
        """ Scales the sequence to fit the sequence """
        firstSelf, lastSelf = self.firstLastOffset()
        durationSelf = lastSelf - firstSelf
        durationScale = lastOffset - firstOffset

        this = self if inplace else self.deepcopy()
        this.addOffset(-firstSelf)
        this.multOffset(durationScale / durationSelf)
        this.addOffset(firstOffset)

        return None if inplace else this

    def normalizeTo(self,
                    aveSv: float = 1.0,
                    inplace: bool = False,
                    ignoreFixed: bool = False,
                    minAllowable: float = None,
                    maxAllowable: float = None,
                    scaleEnd: bool = True) -> SvSequence or None:
        """ Attempts to normalize the whole sequence to a specified average Sv.

        This also respects the fixed Sv concept::

            SVS   1.0 1.5 1.0
            FIXED     [F]
            -----------------
            Normalize to 1.0
            Last SV is ignored
            Hence, We adjust [0:-1]
            However [1] is fixed, so [0] is adjusted.
            -----------------------------------------
            SVS   0.5 1.5 1.0
            FIXED     [F]

        :param aveSv: The SV to average to.
        :param inplace: Whether to perform the operation inplace or not.
        :param ignoreFixed: Whether to ignore the fixed trait and just scale everything.
        :param minAllowable: Minimum allowable SV. Raises AssertionError on failure
        :param maxAllowable: Maximum allowable SV. Raises AssertionError on failure
        :param scaleEnd: Whether to make the end Sv == aveSv
        """

        # Firstly, we find out if it's possible to normalize
        # Last Offset is implicitly the last Sv, which is ignored anyways.
        # noinspection PyTypeChecker
        acts: List[Tuple[SvObj, float]] = activity(self)

        fixedArea: float = 0.0
        looseArea: float = 0.0

        first, last = self.firstLastOffset()
        expectedArea = (last - first) * aveSv

        # Loop through the activities and find the total areas
        for act in acts:
            act: Tuple[SvObj, float]
            if act[0].fixed and not ignoreFixed:  # Check if fixed is ignored also
                fixedArea += act[0].multiplier * act[1]
            else:
                looseArea += act[0].multiplier * act[1]

        requiredScale = (expectedArea - fixedArea) / looseArea

        log.debug(f"Fixed Area: {fixedArea}")
        log.debug(f"Loose Area: {looseArea}")
        log.debug(f"Expected Area: {expectedArea}")
        log.debug(f"Required Loose Scaling: {requiredScale}")

        normSvs: List[SvObj] = []

        # Scale all loose Svs
        for act in acts[:-1]:
            if act[0].fixed and not ignoreFixed:  # Check if fixed is ignored also
                normSvs.append(act[0])
            else:
                sv = act[0]
                sv.multiplier *= requiredScale

                # Assert allowable
                if minAllowable is not None:
                    assert sv.multiplier >= minAllowable, f"Sv Multiplier {sv.multiplier} < Allowable {minAllowable}"
                if maxAllowable is not None:
                    assert sv.multiplier <= maxAllowable, f"Sv Multiplier {sv.multiplier} > Allowable {maxAllowable}"
                normSvs.append(sv)

        # Append last Sv
        if scaleEnd:
            endSv = acts[-1][0]
            endSv.multiplier = aveSv
            normSvs.append(endSv)
        else:
            normSvs.append(acts[-1][0])

        if inplace: self.__init__(normSvs)
        else: return SvSequence(normSvs)

    @overload
    def appendInit(self, list_: List[SvObj]): ...
    @overload
    def appendInit(self, list_: List[Tuple[float, float, bool]]): ...
    @overload
    def appendInit(self, list_: List[Tuple[float, float]]): ...
    @overload
    def appendInit(self, list_: List[float]): ...
    def appendInit(self, list_: List):
        """ Similar to initializing, we can append with multiple ways

        1. `append([SvObj, SvObj, ...])`
        2. `append([(offset, sv), (offset, sv), ...])`
        3. `append([offset, offset, ...])`

        These also can be mixed.

        4. `append([SvObj, (offset, sv), offset, ...])`
        """
        for i in range(len(list_)):
            item = list_[i]
            if isinstance(item, Tuple):
                if len(item) == 2:
                    list_[i] = SvObj(offset=item[0], multiplier=item[1])
                elif len(item) == 3:
                    list_[i] = SvObj(offset=item[0], multiplier=item[1], fixed=item[2])
            elif isinstance(item, float) or isinstance(item, int):
                list_[i] = SvObj(offset=item)
            elif not isinstance(item, SvObj):
                raise TypeError(f"Unexpected Type in list. {item}")

        self.extend(list_)

    def crossWith(self, other: SvSequence, inplace: bool = False):
        """ Cross Multiplies Sequences with each other. Will implicitly sort.

        Consider if you want to have a stutter that slows down, you'd need 2 sequences.

        1. A Slowdown Sequence
        2. A Stutter Sequence

        By crossing them together correctly, you'll achieve a slowdown stutter.

        If you want to keep the Sv Sequence of the stutter only, use `stutter.crossWith(slowdown)`.
        Vice versa.

        If you want to keep everything, including the slowdown during stutter, use fullCrossWith from SvPkg

        For Example::

            self Input         | (1.0) ------- (2.0) ------- (3.0) |
            other Input        | (1.0)  (1.5) ------- (2.0) ------ |
            __________________ | _________________________________ |
            fullCross == False | (1.0) ------- (3.0) ------- (6.0) |
            Not returned       | (1.0)  (1.5) ------- (4.0) ------ |
            fullCross == True  | (1.0)  (1.5)  (3.0)  (4.0)  (6.0) |

        :param other: The Sequence to cross against. Modifies current Sequence
        :param inplace: Whether to perform the operation inplace or not. Affects current Sequence only
        """
        thisI = 0
        otherI = 0
        this = self if inplace else self.deepcopy()
        this.sorted(inplace=True)
        other_ = other.sorted()
        while True:
            if thisI == len(this): break
            thisSv = this[thisI]
            otherSv = other_[otherI]
            multiplier = otherSv.multiplier
            otherNextSv = None if otherI == len(other_) - 1 else other_[otherI + 1]
            if thisSv.offset < otherSv.offset:
                thisI += 1
                continue
            if otherNextSv is not None and thisSv.offset >= otherNextSv.offset:
                otherI += 1
                continue
            thisSv.multiplier *= multiplier
            thisI += 1

        return None if inplace else this

    def multiplyMul(self, by: float, inplace: bool = False):
        """ Multiplies the whole sequence multiplier by

        :param by: The multiplication
        :param inplace: Whether to perform the operation inplace or not
        """
        if inplace:
            for i in self: i.multiplier *= by
        else:
            this = self.deepcopy()
            for i in this: i.multiplier *= by
            return this
    def addMul(self, by: float, inplace: bool = False):
        """ Multiplies the whole sequence multiplier by

        :param by: The addition
        :param inplace: Whether to perform the operation inplace or not
        """
        if inplace:
            for i in self: i.multiplier += by
        else:
            this = self.deepcopy()
            for i in this: i.multiplier += by
            return this

    def __str__(self) -> str:
        return "\n".join(["OFFSET    MULT           FIXED"]
                         + [f"{round(sv.offset,4):<10}"
                            f"{round(sv.multiplier,4):<15}"
                            f"{str(sv.fixed):<7}" for sv in self])




from reamber.osu.OsuBpmObj import OsuBpmObj
from reamber.osu.OsuNoteObjMeta import OsuNoteObjMeta
from reamber.osu.OsuHoldObj import OsuHoldObj
from reamber.osu.OsuSvObj import OsuSvObj
from reamber.osu.OsuHitObj import OsuHitObj
from reamber.osu.OsuSampleSet import OsuSampleSet
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.osu.OsuMapObjMeta import OsuMapObjMeta

__all__ = ['OsuBpmObj', 'OsuHitObj', 'OsuNoteObjMeta', 'OsuHoldObj', 'OsuMapObj', 'OsuMapObjMeta',
           'OsuSampleSet', 'OsuTimingPointMeta', 'OsuSvObj']



import unittest
from tests.test.RSC_PATHS import *

from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj
from reamber.algorithms.analysis.playField import PlayField
from reamber.algorithms.analysis.playField.parts import *

class TestImage(unittest.TestCase):

    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_CARAVAN)
        pf = PlayField(m, padding=70)\
             + PFDrawColumnLines()\
             + PFDrawBeatLines()\
             + PFDrawBpm(xOffset=30, yOffset=0)\
             + PFDrawSv(yOffset=0)\
             + PFDrawNotes()
        pf.exportFold(maxHeight=1000).save("osu.png")

    def test_qua(self):
        m = QuaMapObj()
        m.readFile(QUA_NEURO_CLOUD)
        pf = PlayField(m)\
             + PFDrawColumnLines()\
             + PFDrawBeatLines([1,3,6])\
             + PFDrawNotes()
        pf.exportFold(maxHeight=2000).save("qua.png")

    def test_sm(self):
        s = SMMapSetObj()
        s.readFile(SM_ICFITU)
        pf = PlayField(s.maps[0])\
             + PFDrawBeatLines([1])\
             + PFDrawNotes()
        pf.exportFold(maxHeight=2000).save("sm.png")

    def test_o2j(self):
        s = O2JMapSetObj()
        s.readFile(O2J_FLY_MAGPIE_OJN)
        pf = PlayField(s.maps[2], padding=40)\
             + PFDrawColumnLines()\
             + PFDrawBeatLines([1])\
             + PFDrawBpm()\
             + PFDrawNotes()
        pf.exportFold(maxHeight=2000).save("o2j.png")


if __name__ == '__main__':
    unittest.main()



from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.base.BpmObj import BpmObj
from reamber.sm.SMMapObjMeta import SMMapObjChartTypes
from reamber.sm.SMHitObj import SMHitObj
from reamber.sm.SMHoldObj import SMHoldObj
from reamber.sm.SMBpmObj import SMBpmObj
from reamber.sm.lists.SMBpmList import SMBpmList
from reamber.sm.lists.SMNotePkg import SMNotePkg
from reamber.sm.lists.notes.SMHitList import SMHitList
from reamber.sm.lists.notes.SMHoldList import SMHoldList
from typing import List


class QuaToSM:
    @staticmethod
    def convert(qua: QuaMapObj) -> SMMapSetObj:
        """ Converts a Quaver map to a SMMapset Obj

        Note that each qua map object will create a separate mapset, they are not merged

        :param qua: Quaver map
        :return: SM Mapset
        """
        hits: List[SMHitObj] = []
        holds: List[SMHoldObj] = []

        for hit in qua.notes.hits():
            hits.append(SMHitObj(offset=hit.offset, column=hit.column))
        for hold in qua.notes.holds():
            holds.append(SMHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

        bpms: List[BpmObj] = []

        for bpm in qua.bpms:
            bpms.append(SMBpmObj(offset=bpm.offset, bpm=bpm.bpm))

        smSet: SMMapSetObj = SMMapSetObj(
            music=qua.audioFile,
            title=qua.title,
            titleTranslit=qua.title,
            artist=qua.artist,
            artistTranslit=qua.artist,
            credit=qua.creator,
            background=qua.backgroundFile,
            sampleStart=qua.songPreviewTime,
            sampleLength=10,
            offset=qua.notes.firstOffset(),
            maps=[
                SMMapObj(
                    chartType=SMMapObjChartTypes.DANCE_SINGLE,
                    notes=SMNotePkg(hits=SMHitList(hits),
                                    holds=SMHoldList(holds)),
                    bpms=SMBpmList(bpms)
                )
            ]
        )

        return smSet



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMKeySoundObj import SMKeySoundObj
from typing import List


class SMKeySoundList(List[SMKeySoundObj], SMNoteList):

    def _upcast(self, objList: List = None) -> SMKeySoundList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMKeySoundList
        """
        return SMKeySoundList(objList)

    def data(self) -> List[SMKeySoundObj]:
        return self



""" Takes a chart and constructs an image from it using pillow """

from PIL import Image, ImageDraw, ImageColor
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.sm.SMMapSetObj import SMMapObj
from reamber.o2jam.O2JMapObj import O2JMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from typing import Union

from reamber.algorithms.analysis.playField.parts.PFDrawable import PFDrawable

class PlayField:

    HOLD_RESIZE_BUFFER: int = 2

    def __add__(self, other: PFDrawable):
        assert isinstance(other, PFDrawable), "The added class must be an instance of PFDrawable!"
        return other.draw(pf=self)

    def __init__(self,
                 m: Union[OsuMapObj, O2JMapObj, SMMapObj, QuaMapObj],
                 durationPerPx: float = 5,
                 noteWidth: int = 10,
                 hitHeight: int = 5,
                 holdHeight: int = 5,
                 columnLineWidth: int  = 1,
                 startLead: float = 100.0,
                 endLead: float = 100.0,
                 padding: int = 0,
                 backgroundColor: str = "#000000"):
        """
        Creates an image of the chart

        :param m: The map object
        :param durationPerPx: ms displayed per pixel. The larger the value, the lower the zoom
        :param noteWidth: The width of each note in px
        :param hitHeight: The height of each hit in px
        :param holdHeight: The height of each hold head and tail in px
        :param startLead: The ms lead before the first note. This is used to prevent clipping of the first note
        :param columnLineWidth: The width of column line separating columns
        """

        self.m                = m
        self.durationPerPx    = durationPerPx
        self.noteWidth        = noteWidth
        self.hitHeight        = hitHeight
        self.holdHeight       = holdHeight
        self.columnLineWidth  = columnLineWidth
        self.startLead        = startLead
        self.endLead          = endLead
        self.padding          = padding
        self.backgroundColor  = backgroundColor

        keys = m.notes.maxColumn() + 1

        start, end = m.notes.firstLastOffset()
        start -= startLead
        end += endLead
        duration = end - start

        canvasW = int(noteWidth * keys + columnLineWidth * (keys - 1) + padding)  # -1 due to fencepost
        canvasH = int(duration / durationPerPx)

        canvas = Image.new(mode='RGB', size=(canvasW, canvasH), color=backgroundColor)
        canvasDraw = ImageDraw.Draw(canvas)

        self.keys       = keys
        self.start      = start
        self.end        = end
        self.duration   = duration
        self.canvasH    = canvasH
        self.canvasW    = canvasW
        self.canvas     = canvas
        self.canvasDraw = canvasDraw

    def export(self) -> Image.Image:
        """ Just grabs the image without modifications. I recommend exportFold to make it more squarish """
        return self.canvas

    def exportFold(self,
                   maxHeight: int = 2000,
                   stageLineWidth: int = 3,
                   stageLineColor: str = "#525252") -> Image.Image:
        """ Exports by folding the image

        :param maxHeight: The maximum height of the image, the lower this is, the wider the image
        :param stageLineWidth: The width of the stage line separator
        :param stageLineColor: The color of the stage line separator
        :return:
        """
        # Split the canvas here into stages
        columns = int(self.canvasH / maxHeight + 1)

        newCanvasW = columns * self.canvasW + (columns - 1) * stageLineWidth
        newCanvasH = maxHeight

        newCanvas = Image.new('RGB', (newCanvasW, newCanvasH),
                              color=ImageColor.getrgb(stageLineColor))

        """[SPLIT CANVAS]"""
        for i in range(columns):
            chunk = self.canvas.crop(
                (0                                                      , self.canvasH - (newCanvasH * (i + 1)),
                 self.canvasW + ((i + 1) * self.columnLineWidth) - i - 1, self.canvasH - newCanvasH * i))

            newCanvas.paste(chunk, (i * (self.canvasW + stageLineWidth) , 0))

        # We don't need to draw the column lines because the background already is the column lines

        return newCanvas






import unittest

from reamber.algorithms.generate.sv.generators.svOsuMeasureLineA import svOsuMeasureLineA
from reamber.algorithms.generate.sv.generators.svOsuMeasureLineB import svOsuMeasureLineB
from reamber.osu.OsuBpmObj import OsuBpmObj
from math import sin, pi


class TestMeasureLine(unittest.TestCase):

    def testA(self):

        seq = svOsuMeasureLineA(firstOffset=5000,
                                lastOffset=20000,
                                funcs=[lambda x: 0.5 * sin(x * pi * 2),
                                       lambda x: 0.5 * sin(x * pi * 2 + pi)],
                                fillBpm=200, startX=0, endX=4, endBpm=200, referenceBpm=200,
                                paddingSize=20).combine()

        with open("out.txt", "w+") as f:
            f.writelines([i.writeString() + "\n" for i in seq.writeAsBpm(OsuBpmObj)])

    def testB(self):

        lis = svOsuMeasureLineB(firstOffset=0,
                                lastOffset=40000,
                                funcs=[lambda x: 0.5 * sin(x * pi * 2),
                                       lambda x: 0.5 * sin(x * pi * 2 + pi)],
                                fillBpm=200, startX=0, endX=4, endBpm=100, referenceBpm=200,
                                paddingSize=20)

        with open("out.txt", "w+") as f:
            f.writelines([i.writeString() + "\n" for i in lis])


if __name__ == '__main__':
    unittest.main()



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestRepeat(unittest.TestCase):

    def testRepeat(self):
        # Test Repeat
        seq = SvSequence([SvObj(0, 1.0), SvObj(100, 2.0)])
        seq = SvPkg.repeat(seq, 4).combine(SvPkg.CombineMethod.DROP_BY_POINT)

        self.assertEqual(len(seq), 5)


if __name__ == '__main__':
    unittest.main()



""" Do we still need this ?
Leaving it here just in case

TODO: Decide on if this should be axed"""

from typing import List
from dataclasses import dataclass
from dataclasses import field

from reamber.base.MapObj import MapObj


@dataclass
class MapSetObj:
    """ Deprecated class for mapsets, wasn't sure how this would be used """

    maps: List[MapObj] = field(default_factory=lambda: [])



from reamber.osu.OsuSampleSet import OsuSampleSet
from dataclasses import dataclass


class OsuTimingPointType:
    SV: int = 0
    Bpm: int = 1


@dataclass
class OsuTimingPointMeta:
    """ Holds all metadata for every timing point object"""

    sampleSet: int = OsuSampleSet.AUTO
    sampleSetIndex: int = 0
    volume: int = 50
    kiai: bool = False

    @staticmethod
    def isTimingPoint(s: str) -> bool or None:
        """ Checks whether if the string is a Timing Point/Bpm Obj"""
        t = s.split(",")
        if len(t) < 8:
            return None
        return t[6] == "1"

    @staticmethod
    def isSliderVelocity(s: str) -> bool or None:
        """ Checks whether if the string is a SV Point/SV Obj """
        t = s.split(",")
        if len(t) < 8:
            return None
        return t[6] == "0"




from __future__ import annotations
from reamber.base.lists.notes.NoteList import NoteList
from abc import abstractmethod
from typing import Tuple, List, Dict, Any
import pandas as pd
from dataclasses import asdict
from copy import deepcopy


class NotePkg:
    """ This Package holds multiple note lists """

    @abstractmethod
    def data(self) -> Dict[str, NoteList]:
        """ This grabs the data from inherited instances.
        :rtype: Dict[str, NoteList]
        :return: The inherited instances must return a dictionary of the lists. \
            It is advised to follow the names used in the convention. Such as hit for hits, note for hits and holds.
        """
        ...

    @abstractmethod
    def _upcast(self, dataDict: Dict[str, NoteList]) -> NotePkg:
        """ This just upcasts the current class so that inplace methods can work
        :param dataDict: A dictionary similar to what self.data() outputs
        :rtype: NotePkg
        """
        ...

    def deepcopy(self) -> NotePkg:
        """ Creates a deep copy of itself """
        return deepcopy(self)

    def df(self) -> Dict[str, pd.DataFrame]:
        """ Creates a dict pandas DataFrame by looping through the self.data

        :return: Returns a Dictionary of pd.DataFrames
        """
        # noinspection PyDataclass,PyTypeChecker
        return {key: pd.DataFrame([asdict(obj) for obj in data]) for key, data in self.data().items()}

    def __len__(self) -> int:
        """ Returns the number of lists. For total number of items see objCount() """
        # return sum([len(dataDict) for dataDict in self.data()])
        return len(self.data())

    def objCount(self) -> int:
        """ Returns the total sum number of items in each list. For number of lists use len() """
        return sum([len(data) for data in self.data()])

    def __iter__(self):
        """ Yields the Dictionary item by item """
        yield from self.data()

    def method(self, method: str, **kwargs) -> Dict[str, Any]:
        """ Calls each list's method with eval. Specify method with a string.

        :param method: The method to call, the string must be **EXACT**
        :param kwargs: The extra parameter to use
        :return: Returns a Dict as it may not return a NotePkg init-able
        """
        expression = f"_.{method}(" + ",".join([f"{k}={v}" for k, v in kwargs.items()]) + ")"
        asFunc = eval('lambda _: ' + expression)
        return {key: asFunc(_) for key, _ in self.data().items()}

        # The above is faster for some reason
        # return {key: eval(f"_.{method}(" + ",".join([f"{k}={v}" for k, v in kwargs.items()]) + ")")
        #         for key, _ in self.data().items()}

    def addOffset(self, by, inplace: bool = False) -> NotePkg:
        """ Adds Offset to all items

        :param by: The offset to add, in milliseconds
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.method('addOffset', by=by, inplace=False)
        else: return self._upcast(self.method('addOffset', by=by, inplace=False))

    def inColumns(self, columns: List[int], inplace: bool = False) -> NotePkg:
        """ Filters by columns for all items

        :param columns: The columns to filter by, as a list
        :param inplace: Whether to just modify this instance or return a modified copy
        :return: Returns a modified copy if not inplace
        """
        if inplace: self.method('inColumns', columns=columns, inplace=False)
        else: return self._upcast(self.method('inColumns', columns=columns, inplace=False))

    def columns(self) -> Dict[str, List[int]]:
        """ Gets the columns """
        return self.method('columns')

    def maxColumn(self) -> int:
        """ Gets the maximum column, can be used to determine Key Count if not explicitly stated """
        return max(self.method('maxColumn').values())

    def offsets(self) -> Dict[str, List[float]]:
        """ Gets the offsets """
        return self.method('offsets')

    def firstOffset(self) -> float:
        """ Gets the first offset """
        return min(self.method('firstOffset').values())

    def lastOffset(self) -> float:
        """ Gets the last offset """
        return max(self.method('lastOffset').values())

    def firstLastOffset(self) -> Tuple[float, float]:
        """ Gets the first and last offset, slightly faster because it's only sorted once """
        if len(self.offsets()) == 0: return 0.0, float("inf")
        offsets = sorted([i for j in self.offsets().values() for i in j])  # Flattens the offset list
        return offsets[0], offsets[-1]



from reamber.sm.SMMapSetObjMeta import SMMapSetObjMeta
from reamber.sm.SMRollObj import SMRollObj
from reamber.sm.SMHoldObj import SMHoldObj
from reamber.sm.SMMapSetObj import SMMapSetObj
from reamber.sm.SMMineObj import SMMineObj
from reamber.sm.SMLiftObj import SMLiftObj
from reamber.sm.SMMapObjMeta import SMMapObjMeta
from reamber.sm.SMKeySoundObj import SMKeySoundObj
from reamber.sm.SMFakeObj import SMFakeObj
from reamber.sm.SMStopObj import SMStopObj
from reamber.sm.SMHitObj import SMHitObj
from reamber.sm.SMMapObj import SMMapObj
from reamber.sm.SMBpmObj import SMBpmObj

__all__ = ['SMMapSetObjMeta', 'SMRollObj', 'SMHoldObj', 'SMMapSetObj', 'SMMineObj', 'SMLiftObj',
           'SMMapObjMeta', 'SMKeySoundObj', 'SMFakeObj', 'SMStopObj', 'SMHitObj', 'SMMapObj',
           'SMBpmObj']



import unittest
from tests.test.RSC_PATHS import *

from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj
from reamber.algorithms.analysis.note.nps import npsPlot
import matplotlib.pyplot as plt

class TestNps(unittest.TestCase):

    # @profile
    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_PLANET_SHAPER)
        plt.style.use('dark_background')
        npsPlot(m, binSize=500)

    # # @profile
    # def test_qua(self):
    #     m = QuaMapObj()
    #     m.readFile(QUA_NEURO_CLOUD)
    #     plt.style.use('dark_background')
    #     npsPlot(m, "qua.png", binSize=1000)
    #
    # # @profile
    # def test_sm(self):
    #     s = SMMapSetObj()
    #     s.readFile(SM_CARAVAN)
    #     plt.style.use('dark_background')
    #     npsPlot(s.maps[0], "sm.png")
    #
    # # @profile
    # def test_o2j(self):
    #     s = O2JMapSetObj()
    #     s.readFile(O2J_FLY_MAGPIE_OJN)
    #     plt.style.use('dark_background')
    #     npsPlot(s.maps[2], "o2j.png")


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from reamber.base.TimedObj import TimedObj


@dataclass
class NoteObj(TimedObj):
    """ A Note Object is a playable timed object

    Do not get confused with Hit Object, which is just a single hit/tap.

    The naming convention is done this way to make it clear on what is a note, hit and hold.
    """
    column: int = 0



from reamber.base.lists.notes.NoteList import NoteList
from reamber.base.lists.notes.HoldList import HoldList
from reamber.base.lists.notes.HitList import HitList

__all__ = ['NoteList', 'HoldList', 'HitList']



from __future__ import annotations
from reamber.base.lists.NotePkg import NotePkg
from reamber.quaver.lists.notes.QuaHitList import QuaHitList
from reamber.quaver.lists.notes.QuaHoldList import QuaHoldList
from reamber.quaver.lists.notes.QuaNoteList import QuaNoteList
from typing import Dict, overload


class QuaNotePkg(NotePkg):
    """ This package holds both the hits and holds for each QuaMapObj """

    @overload
    def __init__(self): ...
    @overload
    def __init__(self, dataDict: Dict[str, QuaNoteList]): ...
    @overload
    def __init__(self, hits: QuaHitList, holds: QuaHoldList): ...
    def __init__(self, dataDict=None, hits=None, holds=None):
        """ Initialize a package,

        Can initialize with either overloaded method.

        :param dataDict: The data dictionary, it'll be directly assigned to dataDict. The names must explicitly match
        :param hits: The hits as a QuaHitList
        :param holds: The holds as a QuaHoldList
        """
        if dataDict is not None: self.dataDict = dataDict
        elif hits is not None: self.dataDict = {'hits': hits, 'holds': holds}
        else: self.dataDict: Dict[str, QuaNoteList] = {'hits': QuaHitList(), 'holds': QuaHoldList()}

    def _upcast(self, dataDict: Dict[str, QuaNoteList]) -> QuaNotePkg:
        """ This is to facilitate inherited functions to work """
        return QuaNotePkg(dataDict)

    def __iter__(self):
        """ Yields the Dictionary item by item """
        yield from self.dataDict

    def data(self) -> Dict[str, QuaNoteList]:
        """ Returns the data dictionary of lists """
        return self.dataDict

    # noinspection PyTypeChecker
    def hits(self) -> QuaHitList:
        """ Returns the hitList from the dictionary """
        return self.dataDict['hits']
    # noinspection PyTypeChecker
    def holds(self) -> QuaHoldList:
        """ Returns the holdList from the dictionary """
        return self.dataDict['holds']



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence
from reamber.osu.OsuSvObj import OsuSvObj
from reamber.osu.OsuMapObj import OsuMapObj
from tests.test.RSC_PATHS import *

class TestIO(unittest.TestCase):

    def testSv(self):
        # Complex BPM Points
        osu = OsuMapObj()
        osu.readFile(OSU_CARAVAN)

        seq = SvSequence()
        seq.readSvFromMap(osu)

        for sv, sv0 in zip(seq.writeAsSv(OsuSvObj, volume=0)[:50], osu.svs):
            assert isinstance(sv, OsuSvObj)
            self.assertAlmostEqual(sv0.multiplier, sv.multiplier)
            self.assertEqual(0, sv.volume)

    def testTrueSv(self):
        # Complex BPM Points
        osu = OsuMapObj()
        osu.readFile(OSU_CARAVAN)

        seq = SvSequence()
        seq.readTrueSvFromMap(osu, 140)

        # Just need to check the first 50, others should be ok.
        for sv in seq.writeAsSv(OsuSvObj, volume=0)[:50]:
            assert isinstance(sv, OsuSvObj)
            self.assertAlmostEqual(1, sv.multiplier, delta=0.01)
            self.assertEqual(0, sv.volume)


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from dataclasses import field
from typing import List


class SMMapObjDifficulty:
    BEGINNER: str = "Beginner"
    EASY: str = "Easy"
    MEDIUM: str = "Medium"
    HARD: str = "Hard"
    CHALLENGE: str = "Challenge"
    EDIT: str = "Edit"


class SMMapObjChartTypes:
    # Full Description in CHART_TYPES
    DANCE_SINGLE: str = "dance-single"          # Your normal 4 panel dance mode.
    DANCE_DOUBLE: str = "dance-double"          # Both P1 & P2 pads are used for one player.
    DANCE_SOLO: str = "dance-solo"              # 4-panel, except with additional top-left and top-right columns.
    DANCE_COUPLE: str = "dance-couple"          # One chart, but P1 & P2 have different steps.
    DANCE_THREEPANEL: str = "dance-threepanel"  # Like Single, but the down arrow isn't used.
    DANCE_ROUTINE: str = "dance-routine"        # It's like Couple in that it's for two players
    PUMP_SINGLE: str = "pump-single"            # Single, 5-panel pad.
    PUMP_HALFDOUBLE: str = "pump-halfdouble"    # Uses only six panels in the middle of the pad
    PUMP_DOUBLE: str = "pump-double"            # Same as Dance.
    PUMP_COUPLE: str = "pump-couple"            # Same as Dance.
    PUMP_ROUTINE: str = "pump-routine"          # Same as Dance.
    KB7_SINGLE: str = "kb7-single"              # Standard kb7 layout
    KICKBOX_HUMAN: str = "kickbox-human"        # 4key
    KICKBOX_QUADARM: str = "kickbox-quadarm"    # 4key
    KICKBOX_INSECT: str = "kickbox-insect"      # 6key
    KICKBOX_ARACHNID: str = "kickbox-arachnid"  # 8key
    PARA_SINGLE: str = "para-single"            # 5key.
    BM_SINGLE5: str = "bm-single5"              # 5+1key game mode
    BM_VERSUS5: str = "bm-versus5"              # Unknown, might be the beat equivalent to Couple?
    BM_DOUBLE5: str = "bm-double5"              # Both sides are used.
    BM_SINGLE7: str = "bm-single7"              # 7+1key game mode
    BM_DOUBLE7: str = "bm-double7"              # Both sides are used.
    BM_VERSUS7: str = "bm-versus7"              # Unknown (see versus5)
    EZ2_SINGLE: str = "ez2-single"              # 1 pad
    EZ2_DOUBLE: str = "ez2-double"              # 2 pad
    EZ2_REAL: str = "ez2-real"                  # Divides the hand sensors into upper and lower halves.
    PNM_FIVE: str = "pnm-five"                  # 5key game mode.
    PNM_NINE: str = "pnm-nine"                  # 9key game mode.
    TECHNO_SINGLE4: str = "techno-single4"      # Identical to dance_single
    TECHNO_SINGLE5: str = "techno-single5"      # Identical to pump_single
    TECHNO_SINGLE8: str = "techno-single8"      # eight panels are used: ⬅⬇⬆➡↙↘↗↖
    TECHNO_DOUBLE4: str = "techno-double4"      # Identical to dance_double
    TECHNO_DOUBLE5: str = "techno-double5"      # Identical to pump_double
    TECHNO_DOUBLE8: str = "techno-double8"      # 16 panels (doubles)
    DS3DDX_SINGLE: str = "ds3ddx-single"        # 4key + 4hand...
    MANIAX_SINGLE: str = "maniax-single"        # 4key
    MANIAX_DOUBLE: str = "maniax-double"        # 8key

    @staticmethod
    def getKeys(chart: str) -> int or None:
        if   chart == SMMapObjChartTypes.DANCE_SINGLE:     return 4
        elif chart == SMMapObjChartTypes.DANCE_DOUBLE:     return 8
        elif chart == SMMapObjChartTypes.DANCE_SOLO:       return 6
        elif chart == SMMapObjChartTypes.DANCE_COUPLE:     return 4
        elif chart == SMMapObjChartTypes.DANCE_THREEPANEL: return 3
        elif chart == SMMapObjChartTypes.DANCE_ROUTINE:    return 8
        elif chart == SMMapObjChartTypes.PUMP_SINGLE:      return None
        elif chart == SMMapObjChartTypes.PUMP_HALFDOUBLE:  return None
        elif chart == SMMapObjChartTypes.PUMP_DOUBLE:      return None
        elif chart == SMMapObjChartTypes.PUMP_COUPLE:      return None
        elif chart == SMMapObjChartTypes.PUMP_ROUTINE:     return None
        elif chart == SMMapObjChartTypes.KB7_SINGLE:       return 7
        elif chart == SMMapObjChartTypes.KICKBOX_HUMAN:    return None
        elif chart == SMMapObjChartTypes.KICKBOX_QUADARM:  return None
        elif chart == SMMapObjChartTypes.KICKBOX_INSECT:   return None
        elif chart == SMMapObjChartTypes.KICKBOX_ARACHNID: return None
        elif chart == SMMapObjChartTypes.PARA_SINGLE:      return None
        elif chart == SMMapObjChartTypes.BM_SINGLE5:       return None
        elif chart == SMMapObjChartTypes.BM_VERSUS5:       return None
        elif chart == SMMapObjChartTypes.BM_DOUBLE5:       return None
        elif chart == SMMapObjChartTypes.BM_SINGLE7:       return None
        elif chart == SMMapObjChartTypes.BM_DOUBLE7:       return None
        elif chart == SMMapObjChartTypes.BM_VERSUS7:       return None
        elif chart == SMMapObjChartTypes.EZ2_SINGLE:       return None
        elif chart == SMMapObjChartTypes.EZ2_DOUBLE:       return None
        elif chart == SMMapObjChartTypes.EZ2_REAL:         return None
        elif chart == SMMapObjChartTypes.PNM_FIVE:         return None
        elif chart == SMMapObjChartTypes.PNM_NINE:         return None
        elif chart == SMMapObjChartTypes.TECHNO_SINGLE4:   return None
        elif chart == SMMapObjChartTypes.TECHNO_SINGLE5:   return None
        elif chart == SMMapObjChartTypes.TECHNO_SINGLE8:   return None
        elif chart == SMMapObjChartTypes.TECHNO_DOUBLE4:   return None
        elif chart == SMMapObjChartTypes.TECHNO_DOUBLE5:   return None
        elif chart == SMMapObjChartTypes.TECHNO_DOUBLE8:   return None
        elif chart == SMMapObjChartTypes.DS3DDX_SINGLE:    return None
        elif chart == SMMapObjChartTypes.MANIAX_SINGLE:    return None
        elif chart == SMMapObjChartTypes.MANIAX_DOUBLE:    return None


@dataclass
class SMMapObjMeta:

    chartType: str = SMMapObjChartTypes.DANCE_SINGLE
    description: str = ""
    difficulty: str = SMMapObjDifficulty.EASY
    difficultyVal: int = 1
    grooveRadar: List[float] = field(default_factory=lambda: [0.0, 0.0, 0.0, 0.0, 0.0])

    def _readNoteMetadata(self, metadata: List[str]):
        self.chartType = metadata[0].strip()
        self.description = metadata[1].strip()
        self.difficulty = metadata[2].strip()
        self.difficultyVal = int(metadata[3].strip())
        self.grooveRadar = [float(x) for x in metadata[4].strip().split(",")]

    def _readBpms(self, lines: List[str]):
        pass

    def _readStops(self, lines: List[str]):
        pass



from __future__ import annotations
from typing import List, Tuple
from reamber.base.HoldObj import HoldObj
from abc import ABC, abstractmethod


class HoldList(ABC):
    @abstractmethod
    def data(self) -> List[HoldObj]: ...

    def lastOffset(self) -> float:
        """ Get Last Note Offset """
        return sorted(self.data(), key=lambda x: x.offset)[-1].tailOffset()

    def firstLastOffset(self) -> Tuple[float, float]:
        """ Get First and Last Note Offset
        This is slightly faster than separately calling the singular functions since it sorts once only
        """
        hos = sorted(self.data())
        return hos[0].offset, hos[-1].tailOffset()

    def offsets(self, flatten=True):
        """ Grabs all offsets and tail offsets of objects

        :param flatten: Whether to flatten it to a 1D list
        :return: A 1D List if flatten else 2D
        """
        if flatten: return [i for j in [(obj.offset, obj.tailOffset()) for obj in self.data()] for i in j]
        return [(obj.offset, obj.tailOffset()) for obj in self.data()]

    def tailOffsets(self) -> List[float]:
        return [obj.tailOffset() for obj in self.data()]

    def lengths(self) -> List[float]:
        """ Grabs all object lengths as a list """
        return [obj.length for obj in self.data()]



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMMineObj import SMMineObj
from typing import List


class SMMineList(List[SMMineObj], SMNoteList):

    def _upcast(self, objList: List = None) -> SMMineList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMMineList
        """
        return SMMineList(objList)

    def data(self) -> List[SMMineObj]:
        return self



from reamber.o2jam.O2JMapSetObj import O2JMapSetObj, O2JMapObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.quaver.QuaMapObj import QuaMapObj
from typing import overload


@overload
def mapMetadata(m: O2JMapObj, s: O2JMapSetObj = None, unicode: bool = True) -> str: ...
@overload
def mapMetadata(m: OsuMapObj, s: None = None, unicode: bool = True) -> str: ...
@overload
def mapMetadata(m: QuaMapObj, s: None = None, unicode: bool = True) -> str: ...
@overload
def mapMetadata(m: SMMapObj, s: SMMapSetObj = None, unicode: bool = True) -> str: ...
def mapMetadata(m, s=None, unicode=True) -> str:
    """ Grabs the map metadata

    :param m: The Map Object
    :param s: The Map Set Object, not required for some types
    :param unicode: Whether to try to find the unicode or non-unicode. \
        This doesn't try to convert unicode to ascii, it just looks for if there's an available translation.
    :return:
    """
    def formatting(artist, title, difficulty, creator): return f"{artist} - {title}, {difficulty} ({creator})"
    if isinstance(m, OsuMapObj):
        if unicode: return formatting(m.artistUnicode, m.titleUnicode, m.version, m.creator)
        else: return formatting(m.artist, m.title, m.version, m.creator)

    elif isinstance(m, QuaMapObj):
        return formatting(m.artist, m.title, m.difficultyName, m.creator)

    elif isinstance(m, SMMapObj) and isinstance(s, SMMapSetObj):
        if unicode: return formatting(s.artist if len(s.artist.strip()) > 0 else s.artistTranslit,
                                      s.title if len(s.title.strip()) > 0 else s.titleTranslit,
                                      m.difficulty, s.credit)
        else: return formatting(s.artistTranslit if len(s.artistTranslit.strip()) > 0 else s.artist,
                                s.titleTranslit if len(s.titleTranslit.strip()) > 0 else s.title,
                                m.difficulty, s.credit)

    elif isinstance(m, O2JMapObj) and isinstance(s, O2JMapSetObj):
        try:
            return formatting(s.artist.strip(), s.title, f"Level {s.level[s.maps.index(m)]}", s.creator)
        except IndexError:
            return formatting(s.artist, s.title, "Cannot determine level", s.creator)
        
    else:
        raise NotImplementedError



""" This is only for osu """
from reamber.osu.OsuMapObj import OsuMapObj
import pandas as pd
import numpy as np
import math

from reamber.osu.lists.notes.OsuHitList import OsuHitList, OsuHitObj
from reamber.osu.lists.notes.OsuHoldList import OsuHoldList, OsuHoldObj
from reamber.osu.lists.OsuNotePkg import OsuNotePkg
from reamber.osu.OsuSampleObj import OsuSampleObj

from copy import deepcopy

import logging
log = logging.getLogger(__name__)

def hitSoundCopy(mFrom: OsuMapObj, mTo: OsuMapObj, inplace: bool = False) -> OsuMapObj:
    """ Copies the hitsound from mFrom to mTo
    
    :param inplace: Whether to just modify this instance or return a modified copy
    :param mFrom: The map you want to copy from
    :param mTo: The map you want to copy to, it doesn't mutate this.
    :return: A copy of mTo with the copied hitsounds.
    """
    dfFrom = pd.concat([df for df in mFrom.notes.df().values()], sort=False)
    dfFrom = dfFrom.drop(['column', 'length'], axis='columns', errors='ignore')
    dfFrom = dfFrom[(dfFrom['additionSet'] != 0) |
                    (dfFrom['customSet'] != 0) |
                    (dfFrom['hitsoundSet'] != 0) |
                    (dfFrom['sampleSet'] != 0) |
                    (dfFrom['hitsoundFile'] != "")]
    dfFrom: pd.DataFrame
    dfFrom.sort_values('offset').reset_index(drop=True, inplace=True)

    HITSOUND_CLAP    = 2
    HITSOUND_FINISH  = 4
    HITSOUND_WHISTLE = 8

    # Before we group, we want to split the hitsoundFile to clap, finish and whistle (2, 4, 8)
    dfFrom['hitsoundClap'] \
        = np.where(dfFrom['hitsoundSet'] & HITSOUND_CLAP == HITSOUND_CLAP, HITSOUND_CLAP, 0)
    dfFrom['hitsoundFinish'] \
        = np.where(dfFrom['hitsoundSet'] & HITSOUND_FINISH == HITSOUND_FINISH, HITSOUND_FINISH, 0)
    dfFrom['hitsoundWhistle'] \
        = np.where(dfFrom['hitsoundSet'] & HITSOUND_WHISTLE == HITSOUND_WHISTLE, HITSOUND_WHISTLE, 0)

    dfFrom.drop('hitsoundSet', inplace=True, axis='columns')
    dfFrom = dfFrom.groupby('offset')

    # We'll just get the mTo data then export it again
    dfToNotes = pd.concat(mTo.notes.df(), sort=False)
    dfToNotes.sort_values('offset').reset_index(drop=True, inplace=True)
    dfToOffsets = dfToNotes['offset']

    # We grab a deepCopy if not inplace
    mToCopy = mTo if inplace else deepcopy(mTo)
    mToCopy.resetAllSamples()

    # The idea is to loop through unique offsets where there's hitsounds/samples
    # For each offset, we group by the volume, because we can slot multiple default samples if we just specify 1 volume

    # e.g. < (C)lap (F)inish (W)histle >
    # C F W  Vol | C F W  Vol
    # 1 0 0  20  | 1 1 1  20
    # 0 1 0  20  | 1 0 0  30
    # 0 0 1  20  | 0 1 1  40
    # 1 0 0  30  | CUSTOM 20
    # 0 1 1  40  |
    # CUSTOM 20  |

    for offset, offsetGroup in dfFrom:
        # You cannot have hitsound Files and the default hitsounds together
        # We find out which indexes match on the df we want to copy to
        slotIndexes = list((dfToOffsets == offset)[dfToOffsets == offset].index)
        slot = 0  # Indicates the slot on "TO" we're looking at right now
        slotMax = len(slotIndexes)  # The maximum slots

        offsetGroup: pd.DataFrame
        vGroups = offsetGroup.groupby('volume', as_index=False)\
                             .agg({'hitsoundFile': ';'.join,
                                   'hitsoundClap': 'sum',
                                   'hitsoundFinish': 'sum',
                                   'hitsoundWhistle': 'sum'})
        vGroups: pd.DataFrame

        for k, vGroup in vGroups.iterrows():  # vGroup -> Volume Group
            volume   = vGroup['volume']
            claps    = int(vGroup['hitsoundClap'] / HITSOUND_CLAP)
            finishes = int(vGroup['hitsoundFinish'] / HITSOUND_FINISH)
            whistles = int(vGroup['hitsoundWhistle'] / HITSOUND_WHISTLE)
            hitsoundFiles = [file for file in vGroup['hitsoundFile'].split(';') if len(file) > 0]

            samples = max(claps, finishes, whistles)
            for i in range(0, samples):
                # We loop through the default C F W samples here
                if slot == slotMax:
                    log.debug(f"No slot to place hitsound {slot} > {slotMax}, dropping hitsound at {offset}")
                    break

                val = 0
                if claps:    claps -= 1;    val += HITSOUND_CLAP
                if finishes: finishes -= 1; val += HITSOUND_FINISH
                if whistles: whistles -= 1; val += HITSOUND_WHISTLE
                log.debug(f"Slotted Hitsound {val} at {offset} vol {volume}")
                dfToNotes.at[slotIndexes[slot], 'hitsoundSet'] = val
                dfToNotes.at[slotIndexes[slot], 'volume'] = volume if volume > 0 else 0
                slot += 1

            for file in hitsoundFiles:
                # We loop through the custom sample here
                if slot == slotMax:
                    log.debug(f"No slot to place hitsound {slot} > {slotMax}, sampling {file} at {offset}")
                    mToCopy.samples.append(OsuSampleObj(offset=offset, sampleFile=file, volume=volume))
                    break
                log.debug(f"Slotted Hitsound {file} at {offset} vol {volume}")
                dfToNotes.at[slotIndexes[slot], 'hitsoundFile'] = file
                dfToNotes.at[slotIndexes[slot], 'volume'] = volume if volume > 0 else 0
                slot += 1

    newDf = dfToNotes.to_dict('records')
    newDfHit  = [deepcopy(n) for n in newDf if math.isnan(n['length'])]
    newDfHold = [deepcopy(n) for n in newDf if not math.isnan(n['length'])]
    for n in newDfHit:
        del n['length']
    mToCopy.notes = OsuNotePkg(hits=OsuHitList([OsuHitObj(**hit) for hit in newDfHit]),
                               holds=OsuHoldList([OsuHoldObj(**hold) for hold in newDfHold]))

    return None if inplace else mToCopy



from abc import ABC, abstractmethod


class PFDrawable(ABC):
    """ All PlayField Drawing classes must inherit from this to enable the __add__ op """
    @abstractmethod
    def draw(self, pf: 'PlayField') -> 'PlayField': ...



from __future__ import annotations
from reamber.base.lists.NotePkg import NotePkg
from reamber.o2jam.lists.notes.O2JHitList import O2JHitList
from reamber.o2jam.lists.notes.O2JHoldList import O2JHoldList
from reamber.o2jam.lists.notes.O2JNoteList import O2JNoteList
from typing import Dict, overload


class O2JNotePkg(NotePkg):
    """ This package holds both the hits and holds for each O2JMapObj """

    @overload
    def __init__(self): ...
    @overload
    def __init__(self, dataDict: Dict[str, O2JNoteList]): ...
    @overload
    def __init__(self, hits: O2JHitList, holds: O2JHoldList): ...
    def __init__(self, dataDict=None, hits=None, holds=None):
        """ Initialize a package,

        Can initialize with either overloaded method.

        :param dataDict: The data dictionary, it'll be directly assigned to dataDict. The names must explicitly match
        :param hits: The hits as a O2JHitList
        :param holds: The holds as a O2JHoldList
        """
        if dataDict is not None: self.dataDict = dataDict
        elif hits is not None:   self.dataDict = {'hits': hits, 'holds': holds}
        else: self.dataDict: Dict[str, O2JNoteList] = {'hits': O2JHitList(), 'holds': O2JHoldList()}

    def _upcast(self, dataDict: Dict[str, O2JNoteList]) -> O2JNotePkg:
        """ This is to facilitate inherited functions to work """
        return O2JNotePkg(dataDict)

    def __iter__(self):
        """ Yields the Dictionary item by item """
        yield from self.dataDict

    def data(self) -> Dict[str, O2JNoteList]:
        """ Returns the data dictionary of lists """
        return self.dataDict

    # noinspection PyTypeChecker
    def hits(self) -> O2JHitList:
        """ Returns the hitList from the dictionary """
        return self.dataDict['hits']
    # noinspection PyTypeChecker
    def holds(self) -> O2JHoldList:
        """ Returns the holdList from the dictionary """
        return self.dataDict['holds']



from __future__ import annotations
from reamber.algorithms.analysis.playField import PlayField
from reamber.algorithms.analysis.playField.parts.PFDrawable import PFDrawable

from reamber.osu.OsuMapObj import OsuMapObj
from reamber.quaver.QuaMapObj import QuaMapObj


class PFDrawSv(PFDrawable):

    def __init__(self,
                 decimalPlaces: int = 2,
                 color: str = "#4ef279",
                 xOffset: int = 0,
                 yOffset: int = 0):
        """ Draws Svs on the field, only works with maps that have svs
        :param decimalPlaces: The number of decimal places to display
        :param color: The color of the text
        :param xOffset: Padding from the right, useful if you have multiple text drawables overlapping
        :param yOffset: The offset to move the text
        """
        self.xOffset = xOffset
        self.yOffset = yOffset
        self.decimalPlaces = decimalPlaces
        self.color = color

    def draw(self, pf: PlayField) -> PlayField:
        """ Refer to __init__ """
        assert isinstance(pf.m, OsuMapObj) or isinstance(pf.m, QuaMapObj), "Only sv maps are supported."

        for sv in pf.m.svs:
            txt = f"{sv.multiplier:.{self.decimalPlaces}f}"
            w, h = pf.canvasDraw.textsize(txt)
            x = pf.canvasW - w - self.xOffset
            y = pf.canvasH - int((sv.offset - pf.start) / pf.durationPerPx) - int(h/2) - self.yOffset
            pf.canvasDraw.text(xy=(x, y),
                               text=txt,
                               fill=self.color)

        return pf



import unittest


class MyTestCase(unittest.TestCase):

    """ Not Implemented
    This shouldn't be test-able since it's a base abstract class
    However there's plan to create a universal format, will how that goes.
    For now this is just placeholder
    """

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from reamber.o2jam.lists.notes.O2JNoteList import O2JNoteList
from reamber.o2jam.O2JHoldObj import O2JHoldObj
from reamber.base.lists.notes.HoldList import HoldList
from typing import List


class O2JHoldList(List[O2JHoldObj], O2JNoteList, HoldList):

    def _upcast(self, objList: List = None) -> O2JNoteList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: O2JNoteList
        """
        return O2JHoldList(objList)

    def data(self) -> List[O2JHoldObj]:
        return self



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMHitObj import SMHitObj
from typing import List


class SMHitList(List[SMHitObj], SMNoteList):

    def _upcast(self, objList: List = None) -> SMHitList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMHitList
        """
        return SMHitList(objList)

    def data(self) -> List[SMHitObj]:
        return self



from reamber.quaver.lists.QuaBpmList import QuaBpmList
from reamber.quaver.lists.QuaSvList import QuaSvList
from reamber.quaver.lists.QuaNotePkg import QuaNotePkg

__all__ = ['QuaBpmList', 'QuaNotePkg', 'QuaSvList']



import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.convert.SMToOsu import SMToOsu
from reamber.sm.SMMapSetObj import SMMapSetObj

# import logging
#
# logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestSMToOsu(unittest.TestCase):

    # @profile
    def test_sm1(self):
        # Complex BPM Points

        sm = SMMapSetObj()
        sm.readFile(SM_GRAVITY)

        osuMapSet = SMToOsu.convert(sm)
        osuMapSet[0].audioFileName = "audio.mp3"
        osuMapSet[0].addOffset(15 + 41)
        osuMapSet[0].writeFile("out.osu")

    def test_sm2(self):
        # Stops and multiple map

        sm = SMMapSetObj()
        sm.readFile(SM_ESCAPES)

        osuMapSet = SMToOsu.convert(sm)
        osuMapSet[0].audioFileName = "Escapes.mp3"
        osuMapSet[0].addOffset(635 + 575)
        osuMapSet[0].writeFile("out.osu")

        osuMapSet[1].audioFileName = "Escapes.mp3"
        osuMapSet[1].addOffset(635 + 575)
        osuMapSet[1].writeFile("out.osu")


if __name__ == '__main__':
    unittest.main()



from reamber.base.TimedObj import TimedObj
from dataclasses import dataclass
from typing import Dict


@dataclass
class QuaSvObj(TimedObj):
    multiplier: float = 1.0

    def asDict(self) -> Dict:
        """ Used to facilitate exporting as Qua from YAML """
        return {"StartTime": self.offset,
                "Multiplier": self.multiplier}



from reamber.base.MapObj import MapObj
from reamber.algorithms.analysis.bpm.bpmActivity import bpmActivity

def aveBpm(m: MapObj) -> float:
    """ Calculates the average BPM based on the BPM's Activity on notes

    Used in describe.

    """
    activitySum = 0
    sumProd = 0
    for bpm, activity in bpmActivity(m):
        activitySum += activity
        sumProd += bpm.bpm * activity
    return sumProd / activitySum



from typing import List
from reamber.base.BpmObj import BpmObj
from reamber.base.lists.TimedList import TimedList
from abc import ABC


class BpmList(List[BpmObj], TimedList, ABC):
    """ A List that holds a list of Bpms, useful to do group Bpm operations """

    def data(self) -> List[BpmObj]:
        """ Grabs the list of BpmObj """
        return self

    def bpms(self) -> List[float]:
        """ Grabs a list of Bpm values only """
        return self.attribute('bpm')



from reamber.algorithms.analysis.playField.parts.PFDrawColumnLines import PFDrawColumnLines
from reamber.algorithms.analysis.playField.parts.PFDrawNotes import PFDrawNotes
from reamber.algorithms.analysis.playField.parts.PFDrawBeatLines import PFDrawBeatLines
from reamber.algorithms.analysis.playField.parts.PFDrawBpm import PFDrawBpm
from reamber.algorithms.analysis.playField.parts.PFDrawSv import PFDrawSv

__all__ = ['PFDrawSv', 'PFDrawNotes', 'PFDrawBpm', 'PFDrawColumnLines', 'PFDrawBeatLines']



from __future__ import annotations
from reamber.osu.OsuTimingPointMeta import OsuTimingPointMeta
from reamber.base.TimedObj import TimedObj
from dataclasses import dataclass

MIN_SV = 0.01
MAX_SV = 10.0

@dataclass
class OsuSvObj(OsuTimingPointMeta, TimedObj):
    multiplier: float = 1.0

    @staticmethod
    def codeToValue(code: float) -> float:
        """ Converts the data in the .osu file to the actual SV Value """
        return -100.0 / code

    @staticmethod
    def valueToCode(value: float) -> float:
        """ Converts the actual SV Value to a writable float in .osu """
        return -100.0 / value

    @staticmethod
    def readString(s: str) -> OsuSvObj or None:
        """ Reads a single line under the [TimingPoints] Label. This must explicitly be a BPM Point. """
        if s.isspace():
            return None

        sComma = s.split(",")
        if len(sComma) < 8:
            return None

        this = OsuSvObj()
        assert sComma[6] == '0', "Unexpected BPM Object in OsuSvObj."
        this.offset = float(sComma[0])
        this.multiplier = OsuSvObj.codeToValue(float(sComma[1]))
        this.sampleSet = int(sComma[3])
        this.sampleSetIndex = int(sComma[4])
        this.volume = int(sComma[5])
        this.kiai = int(sComma[7])

        return this

    def writeString(self) -> str:
        """ Exports a .osu writable string """
        return f"{self.offset},{self.valueToCode(self.multiplier)}," \
               f"4,{self.sampleSet}," \
               f"{self.sampleSetIndex},{self.volume},{0},{int(self.kiai)}"



from __future__ import annotations
from reamber.sm.lists.notes.SMNoteList import SMNoteList
from reamber.sm.SMLiftObj import SMLiftObj
from typing import List


class SMLiftList(List[SMLiftObj], SMNoteList):

    def _upcast(self, objList: List = None) -> SMLiftList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: SMLiftList
        """
        return SMLiftList(objList)

    def data(self) -> List[SMLiftObj]:
        return self



from reamber.algorithms.analysis.bpm.bpmActivity import bpmActivity
from reamber.algorithms.analysis.bpm.aveBpm import aveBpm
from reamber.algorithms.analysis.bpm.scrollSpeed import scrollSpeed
from reamber.algorithms.analysis.bpm.bpmBeatOffsets import bpmBeatOffsets

__all__ = ['bpmBeatOffsets', 'scrollSpeed', 'bpmActivity', 'aveBpm']



from reamber.base.BpmObj import BpmObj
from dataclasses import dataclass
from typing import Dict


@dataclass
class QuaBpmObj(BpmObj):
    def asDict(self) -> Dict:
        """ Used to facilitate exporting as Qua from YAML """
        return {
            "StartTime": self.offset,
            "Bpm": self.bpm
        }



import unittest
from reamber.algorithms.generate.sv.SvSequence import SvSequence,SvObj
from reamber.algorithms.generate.sv.SvPkg import SvPkg


class TestAppend(unittest.TestCase):
    def test(self):

        t = SvPkg([SvSequence([SvObj(0, 1.0), SvObj(100, 2.0)]),
                   SvSequence([SvObj(100, 3.0), SvObj(200, 4.0)]),
                   SvSequence([SvObj(150, 5.0), SvObj(300, 6.0)])])\
            .combine(combineMethod=SvPkg.CombineMethod.DROP_BY_POINT, combineMethodWindow=1)

        self.assertEqual(len(t), 5)

        t.appendInit([(0, 1.0), SvObj(100, 2.0), (1000, 2.0, True), 1000])

        self.assertEqual(len(t), 9)


if __name__ == '__main__':
    unittest.main()



from reamber.osu.lists.notes.OsuHoldList import OsuHoldList
from reamber.osu.lists.notes.OsuNoteList import OsuNoteList
from reamber.osu.lists.notes.OsuHitList import OsuHitList

__all__ = ['OsuHoldList', 'OsuNoteList', 'OsuHitList']



from reamber.quaver.lists.notes.QuaNoteList import QuaNoteList
from reamber.quaver.lists.notes.QuaHoldList import QuaHoldList
from reamber.quaver.lists.notes.QuaHitList import QuaHitList

__all__ = ['QuaNoteList', 'QuaHoldList', 'QuaHitList']



from dataclasses import dataclass, field

from reamber.o2jam.O2JEventPackage import O2JEventPackage
from reamber.o2jam.O2JMapSetObjMeta import O2JMapSetObjMeta
from reamber.o2jam.O2JMapObj import O2JMapObj
from typing import List

import logging

log = logging.getLogger(__name__)

@dataclass
class O2JMapSetObj(O2JMapSetObjMeta):
    """ This holds all data of OJN with a few exceptions

    Exceptions:
     - Cover Data
     - Key Sounds Data + Placement

    This also doesn't support OJM (IO) and OJN (O).

    OJM is not supported due to its complexity. OJN writing isn't supported due to lack of support.

    We won't support OJM for now, we'll just deal with OJN since it's much easier. """

    maps: List[O2JMapObj] = field(default_factory=lambda: [])

    def readFile(self, filePath: str):
        """ Reads the OJN file. Do not load the OJM file.

        :param filePath: Path to the ojn file.
        """

        self.__init__()

        with open(filePath, "rb") as f:
            self.readMeta(f.read(300))

            mapPkgs = O2JEventPackage.readEventPackages(f.read(), self.packageCount)
            for pkgs in mapPkgs:
                self.maps.append(O2JMapObj.readPkgs(pkgs=pkgs, initBpm=self.bpm))
            pass

    # def writeFile(self, filePath: str):
    #     with open(filePath, 'wb+') as f:
    #         self.writeMeta(f)



from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
from reamber.o2jam.O2JMapObj import O2JMapObj
from reamber.o2jam.O2JMapSetObjMeta import O2JMapSetObjMeta
from reamber.o2jam.O2JBpmObj import O2JBpmObj
from reamber.o2jam.O2JHoldObj import O2JHoldObj
from reamber.o2jam.O2JHitObj import O2JHitObj
from reamber.o2jam.O2JEventPackage import O2JEventPackage
from reamber.o2jam.O2JNoteObjMeta import O2JNoteObjMeta

__all__ = ['O2JNoteObjMeta', 'O2JEventPackage', 'O2JHitObj', 'O2JMapSetObj', 'O2JMapObj', 'O2JMapSetObjMeta',
           'O2JBpmObj', 'O2JHoldObj']



from __future__ import annotations
from typing import List, Callable, Union
from reamber.algorithms.generate.sv.SvSequence import SvSequence
from enum import Enum
from numpy import linspace
from copy import deepcopy

class SvPkg(List[SvSequence]):

    def __init__(self, list_):
        list.__init__(self, list_)

    class CombineMethod(Enum):
        """ Specifies the combine methods

        - **IGNORE**: Ignores repeated Svs
        - **DROP_BY_POINT**: Drops points that are within window ms of each other
        - **DROP_BY_BOUND**: Drops points that are earlier than the previous list's end.

        Example::

            LIST 1    |[1     1     1]      |
            LIST 2    |   [2  2  2  2  2  2]|
            IGNORE    | 1  2 1&2 2 1&2 2  2 | All combination optimizations are ignored.
            D_B_POINT |[1 [2  1  2  1] 2  2]| Note that the bounds overlap.
            D_B_BOUND |[1     1     1][2  2]| Note that the bounds don't overlap.

        """

        IGNORE = 0
        DROP_BY_POINT = 1
        DROP_BY_BOUND = 2

    def applyNth(self,
                 func: Union[Callable[[float], float], List[float], float],
                 nth: int,
                 startX: float = 0,
                 endX: float = 1) -> None:
        """ Applies the function to the nth element's multiplier of every sequence. Always inplace.

        Example::

            If startX = 0, endX = 1. The function will try to go from 0 to 1 within the sequences linearly.

            If a Callable is passed, if there are 11 sequences, it'll call func(0), func(0.1), ..., func(1).

        :param func: The function to apply. Can be a list of multiplier to apply, or just a constant
        :param nth: The nth element to apply to
        :param startX: The start X
        :param endX: The end X
        """

        if isinstance(func, (float, int)): func = [func for _ in range(len(self))]
        elif isinstance(func, Callable): func = [func(x) for x in list(linspace(startX, endX, len(self)))]

        assert len(func) == len(self), "Lengths must match."
        assert nth < len(self[0]), "nth must be within the list index"

        for mul, seq in zip(func, self):
            seq[nth].multiplier = mul

    def combine(self, combineMethod: CombineMethod = CombineMethod.IGNORE,
                combineMethodWindow: float = 1.0,
                combinePriorityLast: bool = True) -> SvSequence:

        """ Combines multiple sequences together

        Can specify to keep earliest SV if overlapping.

        :param combineMethod: The method to use to combine. See SvSequence.CombineMethod
        :param combineMethodWindow: The millisecond window to check if offsets are of the same offset. Can be 0 for\
            exact comparison
        :param combinePriorityLast: If True, this means that the later SVs will overlap the earlier ones. Recommended\
            for current API
        :return: Returns a stable sorted combine
        """

        if combineMethod == self.CombineMethod.IGNORE:
            return SvSequence([x for y in self for x in y]).sorted(inplace=False)
        elif combineMethod == self.CombineMethod.DROP_BY_POINT:
            newSeq = SvSequence([x for y in self for x in y]).sorted(inplace=False)
            if combinePriorityLast: newSeq.reverse()
            # We loop through the list, if the next offset is similar to current, we delete the next one
            # else we move to the next element
            i = 0
            while i < len(newSeq) - 1:
                if newSeq[i + 1].offset - combineMethodWindow <= newSeq[i].offset <= \
                        newSeq[i + 1].offset + combineMethodWindow:
                    del newSeq[i + 1]
                else:
                    i += 1

            return newSeq.sorted() if combinePriorityLast else newSeq
        else:  # Combine Method == DROP_BY_BOUND
            newSeq = self[0].sorted()
            if combinePriorityLast: newSeq.reverse()
            seqEnd = newSeq.lastOffset()
            for seq in self[1:]:
                addSeq = seq.after(offset=seqEnd, includeEnd=False, inplace=False)
                newSeq += addSeq
                seqEnd = addSeq.lastOffset()
            return newSeq.sorted() if combinePriorityLast else newSeq

    def addOffset(self, by:float, inplace:bool = False) :
        this = self if inplace else deepcopy(self)
        for i in this: i.addOffset(by=by, inplace=True)
        return None if inplace else this

    def multOffset(self, by:float, inplace:bool = False) :
        this = self if inplace else deepcopy(self)
        for i in this: i.multOffset(by=by, inplace=True)
        return None if inplace else this

    @staticmethod
    def fit(seq: SvSequence, offsets: List[float]) -> SvPkg:
        """ Repeats the Sequence such that it repeats from offset to offset, scaled correctly. Always sorts offsets

        Example::

            Input Sequence
            OFFSETS 100 150 200
            SEQ     1.5 0.5 1.0

            Input Offsets
            OFFSETS 100 200 300     500     700

            Output
            SEQ 1        SEQ 2        SEQ 3        SEQ 4
            ---------------------------------------------------
            OFFSET SV  | OFFSET SV  | OFFSET SV  | OFFSET SV  |
            100    1.5 | 200    1.5 | 300    1.5 | 500    1.5 |
            150    0.5 | 250    0.5 | 400    0.5 | 600    0.5 |
            200    1.0 | 300    1.0 | 500    1.0 | 700    1.0 |

        :param seq: The sequence to fit
        :param offsets: The offsets to fit to.
        """
        offsets_ = sorted(offsets)
        seqs = []
        for firstOffset, lastOffset in zip(offsets_[:-1], offsets_[1:]):
            seqs.append(seq.moveStartTo(firstOffset, inplace=False).rescale(firstOffset, lastOffset))

        return SvPkg(seqs)

    @staticmethod
    def repeat(seq: SvSequence, times: int, gap: float = 0) -> SvPkg:
        """ Repeats the Sequence by copying the the sequence to the end.

        Always includes current sequence.

        Consider::

            <---> repeated 3 times.

            <--->
                <--->
                    <--->

        :param seq: The SvSequence To Repeat
        :param times: Number of times to repeat
        :param gap: The gap between each repeat
        """
        first, last = seq.firstLastOffset()
        duration = last - first
        return SvPkg.copyTo(seq=seq, offsets=[first + (duration + gap) * i for i in range(times)])

    @staticmethod
    def copyTo(seq: SvSequence, offsets: List[float]) -> SvPkg:
        """ Copies self to specified offsets

        :param seq: The SvSequence To Copy
        :param offsets: Offsets in float
        :return: Returns a List of SvSequences, flatten-able by SvSequence.combine()
        """
        return SvPkg([seq.deepcopy().addOffset(offset - seq.firstOffset()) for offset in offsets])

    @staticmethod
    def crossMutualWith(this: SvSequence, other: SvSequence) -> SvPkg:
        """ Crosses with each other, returning 2 sequences that can be combined with SvPkg. """
        return SvPkg([this.crossWith(other=other, inplace=False),
                      other.crossWith(other=this, inplace=False)])



from reamber.sm.SMMapSetObjMeta import SMMapSetObjMeta
from reamber.sm.lists.SMBpmList import SMBpmList
from reamber.sm.SMMapObj import SMMapObj
from reamber.sm.SMStopObj import SMStopObj
from dataclasses import dataclass, field
from reamber.sm.SMBpmObj import SMBpmObj

from reamber.base.RAConst import RAConst

from typing import List


@dataclass
class SMMapSetObj(SMMapSetObjMeta):

    maps: List[SMMapObj] = field(default_factory=lambda: [])

    def readFile(self, filePath: str):
        """ Reads a .sm file

        It reads all .sm as a mapset due to the nature of the file format.

        :param filePath: The path to the file
        """
        self.__init__()

        with open(filePath, "r", encoding="utf8") as f:
            file = f.read()
            fileSpl = file.split(";")
            metadata = []
            maps = []
            for index, line in enumerate(fileSpl):
                try:
                    line.index("#NOTES:")
                    maps.append(line)
                except ValueError:
                    metadata.append(line)

            self._readMetadata(metadata)
            bpms = self._readBpms(offset=self.offset, lines=self._bpmsStr)
            self._readStops(lines=self._stopsStr, bpms=bpms)
            self._readMaps(maps=maps, bpms=bpms, stops=self.stops)

            for map in self.maps:
                map.bpms = bpms
            return

    def writeFile(self, filePath: str,
                  alignBpms: bool = False,
                  BEAT_CORRECTION_FACTOR=5.0,
                  BEAT_ERROR_THRESHOLD=0.001):
        """
        Writes the file to filePath specified

        :param BEAT_ERROR_THRESHOLD: See BpmObj.py::alignBpms for details
        :param BEAT_CORRECTION_FACTOR: See BpmObj.py::alignBpms for details
        :param filePath: File Path
        :param alignBpms: Aligns the BPM by mutating the current file. Details in BpmObj.py
        """
        with open(filePath, "w+", encoding="utf8") as f:
            if alignBpms:
                for map in self.maps:
                    map.bpms = SMBpmObj.alignBpms(map.bpms,
                                                     BEAT_CORRECTION_FACTOR=BEAT_CORRECTION_FACTOR,
                                                     BEAT_ERROR_THRESHOLD=BEAT_ERROR_THRESHOLD)
            for s in self._writeMetadata(self.maps[0].bpms):
                f.write(s + "\n")

            for map in self.maps:
                assert isinstance(map, SMMapObj)
                for s in map.writeString():
                    f.write(s + "\n")

    @staticmethod
    def _readBpms(offset: float, lines: List[str]) -> SMBpmList:
        assert offset is not None, "Offset should be defined BEFORE Bpm"
        bpms = []
        beatPrev = 0.0
        bpmPrev = 1.0
        for line in lines:
            beatCurr, bpmCurr = [float(x.strip()) for x in line.split("=")]
            offset += (beatCurr - beatPrev) * RAConst.minToMSec(1.0 / bpmPrev)
            bpms.append(SMBpmObj(offset=offset, bpm=bpmCurr))
            beatPrev = beatCurr
            bpmPrev = bpmCurr

        return SMBpmList(bpms)

    def _readStops(self, bpms: List[SMBpmObj], lines: List[str]):
        for line in lines:
            if len(line) == 0: return
            beatCurr, lengthCurr = [float(x.strip()) for x in line.split("=")]

            index = 0
            for index, bpm in enumerate(bpms):
                if bpm.beat(bpms) > beatCurr:
                    index -= 1
                    break

            offset = bpms[index].offset + (beatCurr - bpms[index].beat(bpms)) * bpms[index].beatLength()

            self.stops.append(SMStopObj(offset=offset, length=RAConst.secToMSec(lengthCurr)))

    def _readMaps(self, maps: List[str], bpms: List[SMBpmObj], stops: List[SMStopObj]):
        for map in maps:
            self.maps.append(SMMapObj.readString(noteStr=map, bpms=bpms, stops=stops))



import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.convert.SMToQua import SMToQua
from reamber.sm.SMMapSetObj import SMMapSetObj

# import logging
#
# logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestSMToQua(unittest.TestCase):

    # @profile
    def test_sm1(self):
        # Complex BPM Points

        sm = SMMapSetObj()
        sm.readFile(SM_GRAVITY)

        quaMapSet = SMToQua.convert(sm)
        quaMapSet[0].writeFile("out.qua")

    def test_sm2(self):
        # Stops and multiple map

        sm = SMMapSetObj()
        sm.readFile(SM_ESCAPES)

        quaMapSet = SMToQua.convert(sm)
        quaMapSet[0].writeFile("out.qua")
        quaMapSet[1].writeFile("out.qua")


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from reamber.base.HitObj import HitObj
from reamber.o2jam.O2JNoteObjMeta import O2JNoteObjMeta


@dataclass
class O2JHitObj(HitObj, O2JNoteObjMeta):
    """ Defines the O2Jam Hit Object

    The O2Jam Hit Object is stored in binary file .ojn
    """
    pass



from reamber.algorithms.generate.sv.SvSequence import SvSequence
from reamber.algorithms.generate.sv.SvPkg import SvPkg
from typing import Callable, List
from reamber.algorithms.generate.sv.generators.svFuncSequencer import svFuncSequencer

from copy import deepcopy

from reamber.osu.OsuBpmObj import MIN_BPM

# The value to use when zero bpm is encountered, using osu's min
FALLBACK_ZERO_BPM = MIN_BPM

def svOsuMeasureLineA(firstOffset: float,
                      lastOffset: float,
                      funcs: List[Callable[[float], float]],
                      referenceBpm: float,
                      endBpm: float or None,
                      paddingSize: int = 10,
                      teleportBpm: float = 1e07,
                      stopBpm: float = 1e-05,
                      fillBpm: float or None = 1e-05,
                      startX: float = 0,
                      endX: float = 1) -> SvPkg:
    """ Generates Measure Line movement for osu! maps. Version 1.

    This is a beta function for svOsuMeasureLine, it may or may not work as expected.

    This handles multi functions a little bit better by stacking them in a single frame instead of flickering through
    them.

    Could be used for other VSRGs but if they support negative Scroll then it could be much easier.

    Sequence::

        S_{_}...F{_F}..._S_T,S_{_}...F{_F}..._S_T,S_{_}...F{_F}..._S_T,...

    :param firstOffset: The first Offset to start the function (x = startX)
    :param lastOffset: The last Offset to end the function (x = endX)
    :param funcs: The functions to use. startX <= x <= endX will be called, expecting a BPM as an output. \
        The more functions you have, the "laggier" it will be.
    :param referenceBpm: The bpm that is used to zero. Found by looking at BPM:XXX-XXX(Reference Bpm) in song select.
    :param paddingSize: The size of the padding, the larger the value, the lower the FPS
    :param teleportBpm: The bpm value for teleporting Bpms.
    :param stopBpm: The bpm value for stop Bpms. Cannot be 0.
    :param fillBpm: The bpm to use to fill such that the sequence ends on lastOffset. None for no fill.
    :param endBpm: The bpm to end the sequence with.
    :param startX: The starting X to use
    :param endX: The ending X to use
    """

    # Optimized value to make sure that 1.0 in input means at the top of the screen.
    # Not accurate for all scrolls and different hit positions.
    SCALING_FACTOR = 9311250 / referenceBpm

    # Append a y = 0 to get diff on first func
    funcs = [lambda x: 0, *funcs]
    funcDiff = []

    for funcI in range(len(funcs) - 1):  # -1 due to the appended y = 0, -1 due to custom last func
        def f(x, i=funcI):
            sort = sorted([g(x) * SCALING_FACTOR for g in funcs])
            for s in range(len(sort)):
                sort[s] = max(0.0, sort[s])  # We eliminate all negative inputs

            out = [g2 - g1 for g1, g2 in zip(sort[:-1], sort[1:])][i]
            if out == 0: return FALLBACK_ZERO_BPM
            else: return out
        funcDiff.append(deepcopy(f))

    funcSeq = []
    funcSeq.extend([stopBpm, *[None for _ in range(paddingSize)], funcDiff[0]])
    for func in funcDiff[1:]: funcSeq.extend([None, func])
    funcSeq.extend([None, stopBpm, None, teleportBpm])

    msecPerFrame = len(funcSeq)

    duration = lastOffset - firstOffset
    frameCount = int(duration / msecPerFrame)

    pkg = svFuncSequencer(funcs=funcSeq,
                          offsets=1,
                          repeats=frameCount,
                          repeatGap=1,
                          startX=startX,
                          endX=endX)

    pkg = SvPkg(map(lambda x: x.addOffset(firstOffset), pkg))

    # Fill missing ending to fit to lastOffset
    if fillBpm is not None:
        seqLastOffset = firstOffset + frameCount * msecPerFrame
        pkg.append(SvSequence([(offset, fillBpm) for offset in range(int(seqLastOffset), int(lastOffset))]))

    if endBpm is not None:
        pkg.append(SvSequence([(lastOffset, endBpm)]))

    return pkg



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List
from reamber.osu.OsuSampleSet import OsuSampleSet
from reamber.osu.OsuSampleObj import OsuSampleObj
from reamber.osu.lists.OsuSampleList import OsuSampleList


class OsuMapObjMode:
    """ This determines the mode of the map.

    Note that only MANIA is supported for now.
    """
    STANDARD: int = 0
    TAIKO: int = 1
    CATCH: int = 2
    MANIA: int = 3


@dataclass
class OsuMapObjMetaGeneral:
    """ All meta under [General] """

    audioFileName: str = ""
    audioLeadIn: int = 0
    previewTime: int = -1
    countdown: bool = False
    sampleSet: int = OsuSampleSet.AUTO
    stackLeniency: float = 0.7
    mode: int = OsuMapObjMode.MANIA
    letterboxInBreaks: bool = False
    specialStyle: bool = False
    widescreenStoryboard: bool = True


@dataclass
class OsuMapObjMetaEditor:
    """ All meta under [Editor] """

    distanceSpacing: float = 4
    beatDivisor: int = 4
    gridSize: int = 8
    timelineZoom: float = 0.3


@dataclass
class OsuMapObjMetaMetadata:
    """ All meta under [Metadata] """

    title: str = ""
    titleUnicode: str = ""
    artist: str = ""
    artistUnicode: str = ""
    creator: str = ""
    version: str = ""
    source: str = ""
    tags: List[str] = ""
    beatmapID: int = 0
    beatmapSetID: int = -1


@dataclass
class OsuMapObjMetaDifficulty:
    """ All meta under [Difficulty] """

    hpDrainRate: float = 5.0
    circleSize: float = 4.0
    overallDifficulty: float = 5.0
    approachRate: float = 5.0
    sliderMultiplier: float = 1.4
    sliderTickRate: int = 1


@dataclass
class OsuMapObjMetaEvents:
    """ All meta under [Events], Excludes Storyboard. """

    backgroundFileName: str = ""
    samples: OsuSampleList = field(default_factory=lambda: OsuSampleList())


@dataclass
class OsuMapObjMeta(OsuMapObjMetaGeneral,
                    OsuMapObjMetaEditor,
                    OsuMapObjMetaMetadata,
                    OsuMapObjMetaDifficulty,
                    OsuMapObjMetaEvents):
    """ The umbrella class that holds everything not included in HitObjects and TimingPoints """

    def readStringList(self, lines: List[str]):
        """ Reads everything Meta """
        for index, line in enumerate(lines):
            if line == "":
                continue

            s = line.split(":")
            if s[0] == "AudioFilename":             self.audioFileName = s[1].strip()
            elif s[0] == "AudioLeadIn":             self.audioLeadIn = int(s[1])
            elif s[0] == "PreviewTime":             self.previewTime = int(s[1])
            elif s[0] == "Countdown":               self.countdown = bool(s[1])
            elif s[0] == "SampleSet":               self.sampleSet = OsuSampleSet.fromString(s[1].strip())
            elif s[0] == "StackLeniency":           self.stackLeniency = float(s[1])
            elif s[0] == "Mode":                    self.mode = int(s[1])
            elif s[0] == "LetterboxInBreaks":       self.letterboxInBreaks = bool(s[1])
            elif s[0] == "SpecialStyle":            self.specialStyle = bool(s[1])
            elif s[0] == "WidescreenStoryboard":    self.widescreenStoryboard = bool(s[1])
            elif s[0] == "DistanceSpacing":         self.distanceSpacing = float(s[1])
            elif s[0] == "BeatDivisor":             self.beatDivisor = int(s[1])
            elif s[0] == "GridSize":                self.gridSize = int(s[1])
            elif s[0] == "TimelineZoom":            self.timelineZoom = float(s[1])
            elif s[0] == "Title":                   self.title = s[1].strip()
            elif s[0] == "TitleUnicode":            self.titleUnicode = s[1].strip()
            elif s[0] == "Artist":                  self.artist = s[1].strip()
            elif s[0] == "ArtistUnicode":           self.artistUnicode = s[1].strip()
            elif s[0] == "Creator":                 self.creator = s[1].strip()
            elif s[0] == "Version":                 self.version = s[1].strip()
            elif s[0] == "Source":                  self.source = s[1].strip()
            elif s[0] == "Tags":                    self.tags = [i.strip() for i in s[1].split(",")]
            elif s[0] == "BeatmapID":               self.beatmapID = int(s[1])
            elif s[0] == "BeatmapSetID":            self.beatmapSetID = int(s[1])
            elif s[0] == "HPDrainRate":             self.hpDrainRate = float(s[1])
            elif s[0] == "CircleSize":              self.circleSize = float(s[1])
            elif s[0] == "OverallDifficulty":       self.overallDifficulty = float(s[1])
            elif s[0] == "ApproachRate":            self.approachRate = float(s[1])
            elif s[0] == "SliderMultiplier":        self.sliderMultiplier = float(s[1])
            elif s[0] == "SliderTickRate":          self.sliderTickRate = int(s[1])

            if s[0] == "//Background and Video events":
                line = lines[index + 1]
                self.backgroundFileName = line[line.find('"')+1:line.rfind('"')]

            if s[0] == "//Storyboard Sound Samples":
                for sampLine in lines[index + 1:]:
                    if not sampLine.startswith('Sample'): break
                    self.samples.append(OsuSampleObj.readString(sampLine))
                break

    def writeStringList(self) -> List[str]:
        """ Writes everything Meta """
        return [
            "osu file format v14",
            "",
            "[General]",
            f"AudioFilename: {self.audioFileName}",
            f"AudioLeadIn: {self.audioLeadIn}",
            f"PreviewTime: {self.previewTime}",
            f"Countdown: {int(self.countdown)}",
            f"SampleSet: {self.sampleSet}",
            f"StackLeniency: {self.stackLeniency}",
            f"Mode: {self.mode}",
            f"LetterboxInBreaks: {int(self.letterboxInBreaks)}",
            f"SpecialStyle: {int(self.specialStyle)}",
            f"WidescreenStoryboard: {int(self.widescreenStoryboard)}",
            "",
            "[Editor]",
            f"DistanceSpacing: {self.distanceSpacing}",
            f"BeatDivisor: {self.beatDivisor}",
            f"GridSize: {self.gridSize}",
            f"TimelineZoom: {self.timelineZoom}",
            "",
            "[Metadata]",
            f"Title:{self.title}",
            f"TitleUnicode:{self.titleUnicode}",
            f"Artist:{self.artist}",
            f"ArtistUnicode:{self.artistUnicode}",
            f"Creator:{self.creator}",
            f"Version:{self.version}",
            f"Source:{self.source}",
            f"Tags:{', '.join(self.tags)}",
            f"BeatmapID:{self.beatmapID}",
            f"BeatmapSetID:{self.beatmapSetID}",
            "",
            "[Difficulty]",
            f"HPDrainRate:{self.hpDrainRate}",
            f"CircleSize:{self.circleSize}",
            f"OverallDifficulty:{self.overallDifficulty}",
            f"ApproachRate:{self.approachRate}",
            f"SliderMultiplier:{self.sliderMultiplier}",
            f"SliderTickRate:{self.sliderTickRate}",
            "",
            "[Events]",
            "//Background and Video events",
            f"0,0,\"{self.backgroundFileName}\",0,0",
            "//Break Periods",
            "//Storyboard Layer 0 (Background)",
            "//Storyboard Layer 1 (Fail)",
            "//Storyboard Layer 2 (Pass)",
            "//Storyboard Layer 3 (Foreground)",
            "//Storyboard Layer 4 (Overlay)",
            "//Storyboard Sound Samples",
            *[sample.writeString() for sample in self.samples]  # Unpacks all samples
        ]



from reamber.base.NoteObj import NoteObj
from dataclasses import dataclass


@dataclass
class HitObj(NoteObj):
    """ A Hit Object is a timed object that is just a single tap

    Do not get confused with Note Object, which describes both hit and holds.
    """
    pass



from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.osu.OsuMapObj import OsuMapObj
from reamber.osu.OsuHitObj import OsuHitObj
from reamber.osu.OsuHoldObj import OsuHoldObj
from reamber.osu.OsuBpmObj import OsuBpmObj
from reamber.base.BpmObj import BpmObj
from reamber.osu.lists.OsuBpmList import OsuBpmList
from reamber.osu.lists.OsuNotePkg import OsuNotePkg
from reamber.osu.lists.notes.OsuHoldList import OsuHoldList
from reamber.osu.lists.notes.OsuHitList import OsuHitList
from typing import List


class SMToOsu:
    @staticmethod
    def convert(sm: SMMapSetObj) -> List[OsuMapObj]:
        """ Converts a SMMapset to possibly multiple osu maps

        Note that a mapset contains maps, so a list would be expected.
        SMMap conversion is not possible due to lack of SMMapset Metadata

        :param sm: The MapSet
        :return: List of Osu Maps
        """

        # I haven't tested with non 4 keys, so it might explode :(

        osuMapSet: List[OsuMapObj] = []
        for smMap in sm.maps:
            assert isinstance(smMap, SMMapObj)

            hits: List[OsuHitObj] = []
            holds: List[OsuHoldObj] = []

            # Note Conversion
            for hit in smMap.notes.hits():
                hits.append(OsuHitObj(offset=hit.offset, column=hit.column))
            for hold in smMap.notes.holds():
                holds.append(OsuHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

            bpms: List[BpmObj] = []

            # Timing Point Conversion
            for bpm in smMap.bpms:
                bpms.append(OsuBpmObj(offset=bpm.offset, bpm=bpm.bpm))

            # Extract Metadata
            osuMap = OsuMapObj(
                backgroundFileName=sm.background,
                title=sm.title,
                titleUnicode=sm.titleTranslit,
                artist=sm.artist,
                artistUnicode=sm.artistTranslit,
                audioFileName=sm.music,
                creator=sm.credit,
                version=f"{smMap.difficulty} {smMap.difficultyVal}",
                previewTime=int(sm.sampleStart),
                bpms=OsuBpmList(bpms),
                notes=OsuNotePkg(hits=OsuHitList(hits),
                                 holds=OsuHoldList(holds))
            )
            osuMapSet.append(osuMap)
        return osuMapSet



""" This holds the metadata for the O2Jam Map Set

This is directly inherited by O2JMapSetObj to allow acsess to all the extra metadata
"""

from dataclasses import dataclass, field
from typing import List
import struct


class O2JMapGenre:
    """ This is a class of static variables that indicate the genre of the song """
    BALLAD     : int = 0
    ROCK       : int = 1
    DANCE      : int = 2
    TECHNO     : int = 3
    HIP_HOP    : int = 4
    SOUL_R_B   : int = 5
    JAZZ       : int = 6
    FUNK       : int = 7
    CLASSICAL  : int = 8
    TRADITIONAL: int = 9
    ETC        : int = 10


@dataclass
class O2JMapSetObjMeta:
    """ This class contains the readable metadata of the map

    This can be extracted from the first 300 bytes of every ojn file."""

    #                                                                # FORMAT   # LENGTH  # STARTS  # ENDS
    songId          : int       = 0                                  # INT      # 4       # 0       # 4
    signature       : str       = ""                                 # CHAR[4]  # 4       # 4       # 8
    encodeVersion   : float     = 0.0                                # FLOAT    # 4       # 8       # 12
    genre           : int       = 0                                  # INT      # 4       # 12      # 16
    bpm             : float     = 0.0                                # FLOAT    # 4       # 16      # 20
    level           : List[int] = field(default_factory=lambda: [])  # SHORT[4] # 8       # 20      # 28
    eventCount      : List[int] = field(default_factory=lambda: [])  # INT[3]   # 12      # 28      # 40
    noteCount       : List[int] = field(default_factory=lambda: [])  # INT[3]   # 12      # 40      # 52
    measureCount    : List[int] = field(default_factory=lambda: [])  # INT[3]   # 12      # 52      # 64
    packageCount    : List[int] = field(default_factory=lambda: [])  # INT[3]   # 12      # 64      # 76
    oldEncodeVersion: int       = 0                                  # SHORT    # 2       # 76      # 78
    oldSongId       : int       = 0                                  # SHORT    # 2       # 78      # 80
    oldGenre        : bytes     = b""                                # CHAR[20] # 20      # 80      # 100
    bmpSize         : int       = 0                                  # INT      # 4       # 100     # 104
    oldFileVersion  : int       = 0                                  # INT      # 4       # 104     # 108
    title           : str       = ""                                 # CHAR[64] # 64      # 108     # 172
    artist          : str       = ""                                 # CHAR[32] # 32      # 172     # 204
    creator         : str       = ""                                 # CHAR[32] # 32      # 204     # 236
    ojmFile         : str       = ""                                 # CHAR[32] # 32      # 236     # 268
    coverSize       : int       = 0                                  # INT      # 4       # 268     # 272
    duration        : List[int] = field(default_factory=lambda: [])  # INT[3]   # 12      # 272     # 284
    noteOffset      : List[int] = field(default_factory=lambda: [])  # INT[3]   # 12      # 284     # 296
    coverOffset     : int       = 0                                  # INT      # 4       # 296     # 300

    BIT_COUNT   = [1,   4,   1,   1,   1,   4,   3,   3,   3,   3,   1,   1,   20,  1,   1,   64,  32,  32,  32,
                   1,   3,   3,   1]
    BIT_SIZES   = [4,   4,   4,   4,   4,   8,   12,  12,  12,  12,  2,   2,   20,  4,   4,   64,  32,  32,  32,
                   4,   12,  12,  4]
    BIT_FORMATS = ["i", "s", "f", "i", "f", "h", "i", "i", "i", "i", "h", "h", "s", "i", "i", "s", "s", "s", "s",
                   "i", "i", "i", "i"]

    def readMeta(self, metadata: bytes):
        """ Reads the metadata of the map

        :param metadata: The first 300 bytes go here
        """
        metaFields: List = []
        indexStart = 0
        for fmt, size, count in zip(O2JMapSetObjMeta.BIT_FORMATS,
                                    O2JMapSetObjMeta.BIT_SIZES,
                                    O2JMapSetObjMeta.BIT_COUNT):
            metaField = []
            size_ = int(size / count)
            for i in range(count):
                metaField.append(struct.unpack("<" + fmt, metadata[indexStart:indexStart + size_])[0])
                indexStart += size_
            metaFields.append(metaField)
        self.songId           = metaFields[0][0]
        self.signature        = b"".join(metaFields[1]).decode("ascii", errors='ignore').replace('\x00', '')
        self.encodeVersion    = metaFields[2][0]
        self.genre            = metaFields[3][0]
        self.bpm              = metaFields[4][0]
        self.level            = metaFields[5]
        self.eventCount       = metaFields[6]
        self.noteCount        = metaFields[7]
        self.measureCount     = metaFields[8]
        self.packageCount     = metaFields[9]
        self.oldEncodeVersion = metaFields[10][0]
        self.oldSongId        = metaFields[11][0]
        self.oldGenre         = b"".join(metaFields[12])
        self.bmpSize          = metaFields[13][0]
        self.oldFileVersion   = metaFields[14][0]
        self.title            = b"".join(metaFields[15]).decode("ascii", errors='ignore').replace('\x00', '')
        self.artist           = b"".join(metaFields[16]).decode("ascii", errors='ignore').replace('\x00', '')
        self.creator          = b"".join(metaFields[17]).decode("ascii", errors='ignore').replace('\x00', '')
        self.ojmFile          = b"".join(metaFields[18]).decode("ascii", errors='ignore').replace('\x00', '')
        self.coverSize        = metaFields[19][0]
        self.duration         = metaFields[20]
        self.noteOffset       = metaFields[21]
        self.coverOffset      = metaFields[22][0]

    def writeMeta(self, f) -> bytes:
        """ Unimplemented, writes the metadata of a ojn file

        I don't think I'll implement this unless there's clear support on this
        """
        pass
        # need to verify all byte sizes on export
        # f.write(struct.pack("<i", self.songId                            ))
        # f.write(bytes(self.signature, encoding='ascii'                   ))
        # f.write(struct.pack("<f", self.encodeVersion                     ))
        # f.write(struct.pack("<i", self.genre                             ))
        # f.write(struct.pack("<f", self.bpm                               ))
        # for level in self.level:
        #     f.write(struct.pack("<h", level                              ))
        # for eventCount in self.eventCount:
        #     f.write(struct.pack("<i", eventCount                         ))
        # for noteCount in self.noteCount:
        #     f.write(struct.pack("<i", noteCount                          ))
        # for measureCount in self.measureCount:
        #     f.write(struct.pack("<i", measureCount                       ))
        # for packageCount in self.packageCount:
        #     f.write(struct.pack("<i", packageCount                       ))
        # f.write(struct.pack("<h", self.oldEncodeVersion                  ))
        # f.write(struct.pack("<h", self.oldSongId                         ))
        # f.write(self.oldGenre                                             )
        # f.write(struct.pack("<i", self.bmpSize                           ))
        # f.write(struct.pack("<i", self.oldFileVersion                    ))
        # # TO#DO: Need to verify length. Fly Magpie is 63, expect 64 bytes
        # f.write(bytes(self.title, encoding='ascii')                       )
        # f.write(struct.pack("<s", bytes(self.artist, encoding='ascii')   ))
        # f.write(struct.pack("<s", bytes(self.creator, encoding='ascii')  ))
        # f.write(struct.pack("<s", bytes(self.ojmFile, encoding='ascii')  ))
        # f.write(struct.pack("<i", self.coverSize                         ))
        # for duration in self.duration:
        #     f.write(struct.pack("<i", self.duration                      ))
        # for noteOffset in self.noteOffset:
        #     f.write(struct.pack("<i", noteOffset                         ))
        # f.write(struct.pack("<i", self.coverOffset                       ))



import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.analysis.bpm.aveBpm import aveBpm
from reamber.osu.OsuMapObj import OsuMapObj

class TestBpmBeatOffsets(unittest.TestCase):

    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_CARAVAN)

        ave = aveBpm(m)


if __name__ == '__main__':
    unittest.main()



from reamber.quaver.QuaNoteObjMeta import QuaNoteObjMeta
from reamber.base.HoldObj import HoldObj
from dataclasses import dataclass
from typing import Dict


@dataclass
class QuaHoldObj(QuaNoteObjMeta, HoldObj):
    def asDict(self) -> Dict:
        """ Used to facilitate exporting as Qua from YAML """
        return {
            'StartTime': self.offset,
            'EndTime': self.offset + self.length,
            'Lane': self.column + 1,
            'KeySounds': self.keySounds
        }



import pandas as pd
from typing import List


def rollingDensity(offsets: List[float], rollingWindowS: float = None) -> pd.Series:
    """ Returns the Density Series for any list

    :param offsets: Any List of offsets
    :param rollingWindowS: The window to search in seconds. If left as None, the window is 0
    :return: Col 0 Offset (DateTime), Col 1 Density (Int)
    """
    df = pd.DataFrame({'offset': offsets})
    df['count'] = 1
    df['offset'] = pd.to_timedelta(df['offset'], unit='ms')
    df = df.groupby('offset').sum()
    if rollingWindowS is not None:
        df = df.rolling(f"{rollingWindowS}s").count()
        df['count'] /= rollingWindowS
        return df.iloc[:, 0]
    else:
        return df.iloc[:, 0]



from __future__ import annotations
from reamber.osu.lists.notes.OsuNoteList import OsuNoteList
from reamber.osu.OsuHoldObj import OsuHoldObj
from reamber.base.lists.notes.HoldList import HoldList
from typing import List


class OsuHoldList(List[OsuHoldObj], OsuNoteList, HoldList):

    def _upcast(self, objList: List = None) -> OsuHoldList:
        """ This is to facilitate inherited functions to work

        :param objList: The List to cast
        :rtype: OsuHoldList
        """
        return OsuHoldList(objList)

    def data(self) -> List[OsuHoldObj]:
        return self



from __future__ import annotations

from reamber.base.MapObj import MapObj
from reamber.base.lists import TimedList
from reamber.sm.SMMapObjMeta import SMMapObjMeta, SMMapObjChartTypes
from reamber.sm.SMBpmObj import SMBpmObj
from reamber.sm.SMStopObj import SMStopObj
from reamber.sm.SMHitObj import SMHitObj
from reamber.sm.SMHoldObj import SMHoldObj
from reamber.sm.SMRollObj import SMRollObj
from reamber.sm.SMMineObj import SMMineObj
from reamber.sm.SMFakeObj import SMFakeObj
from reamber.sm.SMLiftObj import SMLiftObj
from reamber.sm.SMKeySoundObj import SMKeySoundObj

from reamber.sm.SMConst import SMConst

from reamber.sm.lists.SMBpmList import SMBpmList
from reamber.sm.lists.SMNotePkg import SMNotePkg

from dataclasses import dataclass, field
from typing import List, Dict

from numpy import gcd

import logging
log = logging.getLogger(__name__)


@dataclass
class SMMapObj(MapObj, SMMapObjMeta):
    """ If you're trying to load using this, use SMMapSetObj. """

    _SNAP_ERROR_BUFFER = 0.001

    notes: SMNotePkg = field(default_factory=lambda: SMNotePkg())
    bpms:  SMBpmList = field(default_factory=lambda: SMBpmList())

    def data(self) -> Dict[str, TimedList]:
        """ Gets the notes and bpms as a dictionary """
        return {'notes': self.notes,
                'bpms': self.bpms}

    @staticmethod
    def readString(noteStr: str, bpms: List[SMBpmObj], stops: List[SMStopObj]) -> SMMapObj:
        """ Reads the Note part of the SM Map
        That means including the // Comment, and anything below
        :param noteStr: The note part
        :param bpms: BPMs to help sync notes
        :param stops: Stops to help sync notes
        :return:
        """
        spl = noteStr.split(":")
        noteStr = SMMapObj()
        noteStr._readNoteMetadata(spl[1:6])  # These contain the metadata

        # Splits measures by \n and filters out blank + comment entries
        measures: List[List[str]] =\
            [[snap for snap in measure.split("\n")
              if "//" not in snap and len(snap) > 0] for measure in spl[-1].split(",")]

        noteStr._readNotes(measures, bpms=bpms, stops=stops)
        return noteStr

    def writeString(self) -> List[str]:
        """ Write an exportable String List to be passed to SMMapset for writing.
        :return: Exportable String List
        """
        # Tried to use a BPM

        log.info("StepMania writeString is not stable on MultiBpm cases!")
        log.info("Start Parsing File")

        header = [
            f"//------{self.chartType}[{self.difficultyVal} {self.difficulty}]------",
            "#NOTES:",
            f"\t{self.chartType}:",
            f"\t{self.description}:",
            f"\t{self.difficulty}:",
            f"\t{self.difficultyVal}:",
            "\t" + ",".join(map(str, self.grooveRadar)) + ":"
        ]

        log.info(f"Header {header}")

        bpmBeats = SMBpmObj.getBeats(self.bpms, self.bpms)

        # -------- We will grab all required notes here --------
        # List[Tuple[Beat, Column], Char]]
        notes: List[List[float, int, str]] = []

        for snap, ho in zip(SMBpmObj.getBeats(self.notes.hits(), self.bpms), self.notes.hits()):
            notes.append([snap, ho.column, SMConst.HIT_STRING])

        holdObjHeads = []
        holdObjTails = []

        for head, tail in self.notes.holds().sorted().offsets(False):
            holdObjHeads.append(head)
            holdObjTails.append(tail)

        for snap, ho in zip(SMBpmObj.getBeats(holdObjHeads, self.bpms), self.notes.holds()):
            if isinstance(ho, SMHoldObj):   notes.append([snap, ho.column, SMConst.HOLD_STRING_HEAD])
            elif isinstance(ho, SMRollObj): notes.append([snap, ho.column, SMConst.ROLL_STRING_HEAD])

        for snap, ho in zip(SMBpmObj.getBeats(holdObjTails, self.bpms), self.notes.holds()):
            if isinstance(ho, SMHoldObj):   notes.append([snap, ho.column, SMConst.HOLD_STRING_TAIL])
            elif isinstance(ho, SMRollObj): notes.append([snap, ho.column, SMConst.ROLL_STRING_TAIL])

        del holdObjHeads, holdObjTails

        notes.sort(key=lambda x: x[0])

        # -------- Loop through Bpm --------
        # This is where notes are slot into the BPM beats
        # We loop through the BPMs and find which notes fit
        # We then remove the fitted notes and repeat

        # BPM Beat 1                     , BPM Beat 2 ...
        # List[List[Beat, Column, Char]], List[List[Beat, Column, Char]]
        notesByBpm: List[List[float, int, str]] = []
        for bpmBeatIndex in range(len(bpmBeats)):
            # If we are at the end, we use infinity as the upper bound
            bpmBeatLower = bpmBeats[bpmBeatIndex]
            bpmBeatUpper = bpmBeats[bpmBeatIndex + 1] if bpmBeatIndex < len(bpmBeats) - 1 else float("inf")

            # Filter out placement for this bpm beat
            noteByBpm: List[List[float, int, str]] = []
            noteIndexToRemove = []
            for noteIndex, note in enumerate(notes):
                # We exclude the any notes are that close to the lower BPM Beat else they will repeat
                if bpmBeatLower - self._SNAP_ERROR_BUFFER <= note[0] < bpmBeatUpper + self._SNAP_ERROR_BUFFER:
                    log.info(f"Write Note: Beat {round(note[0], 2)}, Column {note[1]}, Char {note[2]} set in "
                             f"{round(bpmBeatLower, 1)} - {round(bpmBeatUpper, 1)}")
                    noteByBpm.append(note)
                    noteIndexToRemove.append(noteIndex)

            # Remove filtered out objects
            noteIndexToRemove.reverse()  # We need to reverse the list to retain correct indexes
            for index in noteIndexToRemove:
                del notes[index]  # faster than pop

            # Zeros the measure and converts it into snap units
            noteByBpm = [[round(m * 48), c, ch] for m, c, ch in noteByBpm]
            notesByBpm += noteByBpm

        del noteByBpm, notes, bpmBeatIndex, bpmBeatUpper, bpmBeatLower, note, noteIndexToRemove, index

        notesByBpm.sort(key=lambda item: item[0])

        # -------- Fit into Measures --------
        # After finding which notes belong to which BPM
        # We cut them into measures then slot them in
        # Note that we want to have the smallest size template before slotting
        # That's where GCD comes in handy.

        measures = [[] for _ in range(int(notesByBpm[-1][0] / 192) + 1)]
        keys = SMMapObjChartTypes.getKeys(self.chartType)
        for note in notesByBpm:
            measures[int(note[0] / 192)].append(note)

        measuresStr = []
        for measureIndex, measure in enumerate(measures):
            log.info(f"Parse Measure {measureIndex}\t{measure}")
            measure = [[snap % 192, col, char] for snap, col, char in measure]
            log.info(f"Zero Measure\t\t{measure}")
            if len(measure) != 0:
                # Using GCD, we can determine the smallest template to use
                gcd_ = gcd.reduce([x[0] for x in measure])
                if gcd_ == 0: snapsReq: int = 4
                else: snapsReq: int = int(192 / gcd_)

                log.info(f"Calc Snaps Req.\t{int(snapsReq)}")
                if snapsReq == 3: snapsReq = 6  # Min 6 snaps to represent
                if snapsReq < 4: snapsReq = 4  # Min 4 snaps to represent
                log.info(f"Final Snaps Req.\t{int(snapsReq)}")

                # This the template created to slot in notes
                measure = [[int(snap/(192/snapsReq)), col, char] for snap, col, char in measure]

                measureStr = [['0' for _key in range(keys)] for _snaps in range(int(snapsReq))]
                log.info(f"Write Measure Input \t\t{measure}")

                # Note: [Snap, Column, Char]
                for note in measure: measureStr[note[0]][note[1]] = note[2]
            else:
                measureStr = [['0' for _key in range(keys)] for _snaps in range(4)]
            measuresStr.append("\n".join(["".join(snap) for snap in measureStr]))
            log.info(f"Finished Parsing Measure")

        log.info(f"Finished Parsing Notes")
        return header + ["\n,\n".join(measuresStr)] + [";\n\n"]

    def _readNotes(self, measures: List[List[str]], bpms: List[SMBpmObj], stops: List[SMStopObj]):
        """ Reads notes from split measures
        We expect a format of [['0000',...]['0100',...]]
        :param measures: Measures as 2D List
        :param bpms: BPMs to help sync
        :param stops: Stops to help Sync
        """
        globalBeatIndex: float = 0.0  # This will help sync the bpm used
        currentBpmIndex: int = 0
        currentStopIndex: int = -1
        offset = bpms[0].offset
        stopOffsetSum = 0

        bpmBeats = SMBpmObj.getBeats(bpms, bpms)
        stopBeats = SMBpmObj.getBeats(stops, bpms)

        # The buffer is used to find the head and tails
        # If we find the head, we throw it in here {Col, HeadOffset}
        # If we find the tail, we extract ^ and clear the Dict then form the Hold/Roll
        holdBuffer: Dict[int, float] = {}
        rollBuffer: Dict[int, float] = {}

        for measure in measures:
            for beatIndex in range(4):
                # Grabs the first beat in the measure
                beat = measure[int(beatIndex * len(measure) / 4): int((beatIndex + 1) * len(measure) / 4)]

                # Loop through the beat
                for snapIndex, snap in enumerate(beat):
                    for columnIndex, columnChar in enumerate(snap):
                        # "Switch" statement for character found
                        if columnChar == "0":
                            continue
                        elif columnChar == SMConst.HIT_STRING:
                            self.notes.hits().append(SMHitObj(offset + stopOffsetSum, column=columnIndex))
                            log.info(f"Read Hit at \t\t{round(offset + stopOffsetSum)} "
                                     f"at Column {columnIndex}")
                        elif columnChar == SMConst.MINE_STRING:
                            self.notes.hits().append(SMMineObj(offset + stopOffsetSum, column=columnIndex))
                            log.info(f"Read Mine at \t\t{round(offset + stopOffsetSum, 2)} "
                                     f"at Column {columnIndex}")
                        elif columnChar == SMConst.HOLD_STRING_HEAD:
                            holdBuffer[columnIndex] = offset
                            log.info(f"Read HoldHead at \t{round(offset + stopOffsetSum, 2)} "
                                     f"at Column {columnIndex}")
                        elif columnChar == SMConst.ROLL_STRING_HEAD:
                            rollBuffer[columnIndex] = offset
                            log.info(f"Read RollHead at \t{round(offset + stopOffsetSum, 2)} "
                                     f"at Column {columnIndex}")
                        elif columnChar == SMConst.ROLL_STRING_TAIL:  # ROLL and HOLD tail is the same
                            #  Flush out hold/roll buffer
                            if columnIndex in holdBuffer.keys():
                                startOffset = holdBuffer.pop(columnIndex)
                                self.notes.holds().append(SMHoldObj(startOffset + stopOffsetSum,
                                                                     column=columnIndex,
                                                                     length=offset - startOffset))
                                log.info(f"Read HoldTail at \t{round(startOffset + stopOffsetSum, 2)} "
                                         f"of length {round(offset - startOffset, 2)} "
                                         f"at Column {columnIndex}")
                            elif columnIndex in rollBuffer.keys():
                                startOffset = rollBuffer.pop(columnIndex)
                                self.notes.holds().append(SMRollObj(startOffset + stopOffsetSum,
                                                                     column=columnIndex,
                                                                     length=offset - startOffset))
                                log.info(f"Read RollTail at \t{round(startOffset + stopOffsetSum, 2)} "
                                         f"of length {round(offset - startOffset, 2)} "
                                         f"at Column {columnIndex}")
                        elif columnChar == SMConst.LIFT_STRING:
                            self.notes.hits().append(SMLiftObj(offset=offset + stopOffsetSum,
                                                                column=columnIndex))
                            log.info(f"Read Lift at \t\t{round(offset + stopOffsetSum, 2)} "
                                     f"at Column {columnIndex}")
                        elif columnChar == SMConst.FAKE_STRING:
                            self.notes.hits().append(SMFakeObj(offset=offset + stopOffsetSum,
                                                                column=columnIndex))
                            log.info(f"Read Fake at \t\t{round(offset + stopOffsetSum, 2)} "
                                     f"at Column {columnIndex}")
                        elif columnChar == SMConst.KEYSOUND_STRING:
                            self.notes.hits().append(SMKeySoundObj(offset=offset + stopOffsetSum,
                                                                    column=columnIndex))
                            log.info(f"Read KeySound at \t{round(offset + stopOffsetSum, 2)} "
                                     f"at Column {columnIndex}")

                    globalBeatIndex += 4.0 / len(measure)
                    offset += bpms[currentBpmIndex].beatLength() / len(beat)
                    #         <-  Fraction  ->   <-    Length of Beat     ->
                    #         Length of Snap

                    # Check if next index exists & check if current beat index is outdated
                    while currentBpmIndex + 1 != len(bpms) and \
                            globalBeatIndex > bpmBeats[currentBpmIndex + 1] - self._SNAP_ERROR_BUFFER:
                        globalBeatIndex = bpmBeats[currentBpmIndex + 1]
                        currentBpmIndex += 1

                    # Add stop offsets to current offset sum
                    while currentStopIndex + 1 != len(stops) and \
                            globalBeatIndex > stopBeats[currentStopIndex + 1]:
                        stopOffsetSum += stops[currentStopIndex + 1].length
                        currentStopIndex += 1

                # Deal with rounding issues
                globalBeatIndex = round(globalBeatIndex)



from reamber.o2jam.lists.notes.O2JHoldList import O2JHoldList
from reamber.o2jam.lists.notes.O2JNoteList import O2JNoteList
from reamber.o2jam.lists.notes.O2JHitList import O2JHitList

__all__ = ['O2JHoldList', 'O2JNoteList', 'O2JHitList']



from reamber.base.lists.NotePkg import NotePkg
from reamber.base.lists.BpmList import BpmList
from abc import ABC, abstractmethod
from typing import Dict
from reamber.base.lists.TimedList import TimedList


class MapObj(ABC):
    """ This class should be inherited by all Map Objects

    They must inherit the data method, which extracts all data they hold.
    They are also assumed to be a TimedList.
    """

    notes: NotePkg
    bpms: BpmList

    @abstractmethod
    def data(self) -> Dict[str, TimedList]:
        """ Gets the data as a dictionary """
        ...

    def addOffset(self, by: float):
        """ Move all by a specific ms """

        for k, i in self.data().items():
            i.addOffset(by)


import unittest
from tests.test.RSC_PATHS import *

from reamber.osu.OsuMapObj import OsuMapObj
from reamber.algorithms.analysis.bpm.bpmBeatOffsets import bpmBeatOffsets

class TestBpmBeatOffsets(unittest.TestCase):

    # @profile
    def test_osu(self):
        m = OsuMapObj()
        m.readFile(OSU_CARAVAN)
        bpmBeatOffsets(m.bpms, 4, 10000)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from reamber.base.TimedObj import TimedObj as TimedObj
from reamber.base.RAConst import RAConst
from dataclasses import dataclass

from typing import List
from typing import Tuple
from typing import Union


@dataclass
class BpmObj(TimedObj):
    """ A non-playable timed object that specifies the tempo of the map.

    This is synonymous with Bpm Point, it's named Object to make it consistent
    """

    bpm: float = 120.0
    metronome: int = 4

    def beatLength(self) -> float:
        """ Returns the length of the beat in ms """
        return RAConst.minToMSec(1.0 / self.bpm)

    def metronomeLength(self) -> float:
        """ Returns the length of the beat in metronome """
        return self.beatLength() * self.metronome

    def beat(self, bpms: List[BpmObj]):
        """ Gets the beat of the current BPM Point w.r.t. bpms """
        return BpmObj.getBeats([self.offset], bpms)[0]

    @staticmethod
    def snapExact(offsets: List[float], bpms: List[BpmObj], snapPrecision: int = 64):
        """ Snaps the offsets to the exact snap

        Returns in the exact same order.

        Example::

            BpmObj.snapExact([1, 100, 250, 385], bpms=[BpmObj(0, 150)], snapPrecision=16)
            [400.0, 300.0, 100.0, 0.0]

        """
        offsetsSort = sorted(offsets, reverse=True)
        offsetsIndex = [offsets.index(x) for x in offsetsSort]

        bpmI = 0
        bpmsSorted = sorted(bpms, reverse=True, key=lambda x: x.offset)
        offsetsOut = []

        for offset in offsetsSort:
            try:
                while offset < bpmsSorted[bpmI].offset:
                    bpmI += 1
            except IndexError:
                raise IndexError("Offset located before first BPM")
            snapLength = RAConst.minToMSec(4 / (snapPrecision * bpmsSorted[bpmI].bpm))
            error = (offset - bpmsSorted[bpmI].offset) % snapLength
            if error < snapLength / 2: offsetsOut.append(offset - error)
            else:                      offsetsOut.append(offset + (snapLength - error))

        # This returns the offsets sorted as the original
        return [x for _,x  in sorted(zip(offsetsIndex, offsetsOut), key=lambda x: x[0])]

    # Beats are used for the StepMania format
    # One large caveat for beats is that it requires looping through the BPM Points of the map to
    # calculate 1 single beat
    # Hence There is a getBeats and getBeat, where if getBeats is used for a list instead of looping through
    # getBeat, it'll be marginally faster.

    @staticmethod
    def getBeats(offsets: Union[List[float], List[TimedObj], float],
                 bpms: List[BpmObj]) -> List[float]:
        """ Gets the beat numbers from offsets provided, this is relative to the first Timing Point

        :param offsets: Offsets to find beat from, can be a list of TOs or floats or a single float
        :param bpms: The Global BPM list
        :return: Beat numbers with respect to the first Timing Point provided.
        """
        # The idea here is to loop through the BPM Points once while filling the beats list with the correct offsets
        # 1.  For each offset:
        # 1.1 Go through the bpms until we find the 2 BPMs that sandwich it
        #     [BPM] [Offset] [BPM]
        # 1.2 If the offset cannot be sandwiched, then we use the last bpm to reference
        #     [BPM] [Offset]
        #
        # 2.  While we are looping through BPMs to find, we need to take note of the prior Beat
        # 2.1 For each BPM iteration, we update bpmPrevBeat
        #
        # 3.  Once we found the sandwich, we find the difference in offsets and find out the beat of the offset

        # Main problem obviously is if the offsets or bpms are not sorted, it'll break the algorithm
        # So we need to get a sorted offset and sorted bpm
        # Unfortunately, unless we force implement sorting on the higher level,
        # this sorting algorithm will have to stay

        # Firstly, coerce the offsets into a List[float] if it's not already
        offsets_ = offsets if isinstance(offsets, List) else [offsets]
        offsets_: Union[List[TimedObj], List[float]]
        if len(offsets_) == 0: return []
        offsets_ = [x.offset for x in offsets_] if isinstance(offsets_[0], TimedObj) else offsets_
        offsets_: List[float]

        # We attach an enum to the original list and sort by the offsets, this sorts it once
        # noinspection PyTypeChecker
        offsetsSorted_: List[Tuple[int, float]] = [x for x in sorted(enumerate(offsets_), key=lambda x:x[1])]
        offsetsSortedOrder: List[int]  # This is the original order
        offsetsSorted: List[float]
        offsetsSortedOrder, offsetsSorted = zip(*offsetsSorted_)  # Unpacks into the original order and sort

        del offsetsSorted_, offsets_

        bpmIndex = 0
        bpmPrevBeat = 0  # If we skip TPs, we have to account for their previous beat

        beats: List[float] = []
        bpms.sort(key=lambda x: x.offset)

        for offset in offsetsSorted:
            # Shift TP such that tp is the latest
            while bpmIndex != len(bpms) - 1 and \
                    not (bpms[bpmIndex].offset <= offset < bpms[bpmIndex + 1].offset):
                bpmIndex += 1
                bpmPrevBeat += (bpms[bpmIndex].offset - bpms[bpmIndex - 1].offset) /\
                                bpms[bpmIndex - 1].beatLength()

            beats.append(bpmPrevBeat + ((offset - bpms[bpmIndex].offset) / bpms[bpmIndex].beatLength()))

        # Sorts beats by original order
        return [x for x, _ in sorted(zip(beats, offsetsSortedOrder), key=lambda x:x[1])]

    @staticmethod
    def alignBpms(bpms: List[BpmObj],
                  BEAT_ERROR_THRESHOLD: float = 0.001,
                  BEAT_CORRECTION_FACTOR: float = 5.0) -> List[BpmObj]:
        """ Ensures that all BPMs are on an integer measure by adding or amending

        :param bpms: The BPMs
        :param BEAT_ERROR_THRESHOLD: If the fraction change in beat's is more than this, we append a new BPM, else we \
            create a new bpm. i.e. if there's too many high BPM points, then increase this
        :param BEAT_CORRECTION_FACTOR: The number of beats to search prior to the affected beat to amend. i.e. if \
            there are too many incorrectly snapped notes DUE TO ADDED BPMs then decrease this
        :return: The new and old BPMs
        """
        # Summary
        # We ensure that all BPMs are on an integer measure, this makes it such that it's easier to snap

        # This will be a conditional of alignBpms
        # If the BPM is misaligned to a measure
        #   If there is a BPM less than or equal to a beat prior
        #       Adjust that one
        #   Else
        #       Create a BPM 1 Beat prior

        # Edge cases
        # Sometimes the beat is like X.000001
        # For this, we wouldn't want to add a BPM at X and shift this to X+1
        # This is because we'll get a very high BPM here and it'll look ugly
        # For this, we will add a BPM at X-1 or amend X-1.

        # Hence
        # If X.000        : Don't change anything
        # If Amend        : Just amend the previous
        # If X.000 ~ X.001: Change X-1
        # If X.001 ~ X.999: Change X   & Move to X + 1
        # BEAT_ERROR_THRESHOLD

        # The beat correction factor defines how many beats to look behind to correct a bpm point
        # BEAT_CORRECTION_FACTOR

        bpmBeats = BpmObj.getBeats(bpms, bpms)
        bpmsNew: List[BpmObj] = [bpms[0]]
        for bpmIndex in range(1, len(bpms)):  # Note We don't touch the first bpm, that's assumed to be 0.0
            bpmPrev = bpms[bpmIndex - 1]
            bpmCurr = bpms[bpmIndex]
            bpmBeatPrev = bpmBeats[bpmIndex - 1]
            bpmBeatCurr = bpmBeats[bpmIndex]

            bpmBeatError = (bpmBeatCurr - bpmBeatPrev) % 1.0

            if bpmBeatError == 0.0:
                pass
            elif bpmCurr.offset - bpmCurr.beatLength() * BEAT_CORRECTION_FACTOR <= bpmPrev.offset <= bpmCurr.offset:
                # This is the case when the previous BPM is within (1 * BCF) beats of the current BPM
                # Instead of inserting another BPM, we amend the previous BPM
                bpmsNew[-1].bpm = 1 / RAConst.mSecToMin(bpmCurr.offset - bpmPrev.offset)
            elif bpmBeatError < BEAT_ERROR_THRESHOLD:
                # As defined, if we happen to have an error that's less than the threshold, instead of forcing
                # a new bpm, we amend the prior bpm OR add a bpm point 1 beat before
                bpmsNew.append(BpmObj(offset=bpmCurr.offset - (1.0 + bpmBeatError) * bpmPrev.beatLength(),
                                         bpm=1.0 / RAConst.mSecToMin(bpmPrev.beatLength() * (1 + bpmBeatError))))
            else:
                # This is the case when the beat is significant enough that it warrants a BPM point in its place
                bpmsNew.append(BpmObj(offset=bpmCurr.offset - bpmBeatError * bpmPrev.beatLength(),
                                         bpm=1.0 / RAConst.mSecToMin(bpmPrev.beatLength() * bpmBeatError)))

            bpmsNew.append(bpmCurr)

        return bpmsNew

    # This is the previous method to alignBpms, it's not very good haha...
    # @staticmethod
    # def alignBpms(bpmPoints: List[BpmObj]) -> List[BpmObj]:
    #     # The naive approach is to forcibly link all bpmBeats together
    #     # We do this by creating a BPM 1ms before all incorrectly offset BPMs and push that BPM forward to an integer
    #     # E.g.
    #     # [1, 2.5, 3.5, 5] -> [1, 2.5 - e, 3, 4.5, 6] -> [1, 2.5 - e, 3, 4.5 - e, 5, 7]
    #     # where e is the beat snap for 1 ms, it'll be based on the previous BPM
    #
    #     # Formula
    #     # BPM: 60000 - mod * 60000 + prevBPM
    #     # Offset: currBpm.offset - prevBpm/60000
    #
    #     bpmPointsSorted = sorted(bpmPoints, key=lambda x: x.offset)
    #     bpmBeats = BpmObj.getBeats(bpmPointsSorted, bpmPointsSorted)
    #     newBpms: List[BpmObj] = [bpmPointsSorted[0]]
    #
    #     # We naively assume that all bpms are incorrect except the first
    #     for bpmIndex, (bpm, bpmBeat) in enumerate(zip(bpmPointsSorted[1:], bpmBeats[1:])):
    #         beatShift: float = bpmBeat % 1.0
    #
    #         # This means that if the measure reset is within 1/4, 16th of a beat, and has an error of < 0.001, we
    #         # will adjust it, else ignore
    #         if beatShift % 0.125 < 0.001 or beatShift % 0.125 > 0.124:
    #             newBpms.append(bpm)
    #             continue
    #
    #         beatShift = beatShift if 0.0 < beatShift <= 1.0 else 1.0
    #
    #         # # Correction BPM (4th) < Refer to 16ths >
    #         # # This would be a large change in BPM
    #         # newBpms.append(BpmObj(
    #         #     # This will shift the offset to the closest prev integer beat
    #         #     offset=bpm.offset - beatShift * bpmPoints[bpmIndex].beatLength(),
    #         #     bpm=1 / RAConst.mSecToMin(bpmPoints[bpmIndex].beatLength() * beatShift)))
    #
    #         # Correction BPM (16th)
    #         # The idea here would be a bit different, we'll have the correction and shifted BPM on integers.
    #         # Let's say we have the beatShift as 0.333
    #         # Then we adjust the Correction to 0.0, the Shift to 1.0, we'll then calculate the required BPM
    #         # Side-effect would be that notes may have problems syncing to the new BPM, we'll deal with that
    #         # later
    #         newBpms.append(BpmObj(
    #             # This will shift the offset to the closest prev integer beat
    #             offset=bpm.offset - beatShift * bpmPoints[bpmIndex].beatLength(),
    #             bpm=1 / RAConst.mSecToMin(bpmPoints[bpmIndex].beatLength() * beatShift)))
    #
    #         # Correction BPM (192nd)
    #         # newBpms.append(BpmObj(offset=bpm.offset - bpmPoints[bpmIndex].beatLength() / 48,  # 1/192
    #         #                         bpm=(1 - beatOffset + 1 / 48) /
    #         #                             (RAConst.mSecToMin(bpmPoints[bpmIndex].beatLength()) / 48)))
    #
    #         # Correction BPM (1ms)
    #         # newBpms.append(BpmObj(offset=bpm.offset - 1,
    #         #                         bpm=60000 - beatShift * 60000 + bpmPoints[bpmIndex].bpm))
    #
    #         # Shifted BPM
    #         newBpms.append(bpm)
    #
    #     # This removes all repeating offsets, priority goes to the new BPMs
    #     bpmIndexToRemove = []
    #     for bpmIndex in range(len(newBpms) - 1):
    #         if newBpms[bpmIndex + 1].offset == newBpms[bpmIndex].offset:
    #             bpmIndexToRemove.append(bpmIndex)
    #
    #     bpmIndexToRemove.reverse()
    #     for bpmIndex in bpmIndexToRemove:
    #         newBpms.pop(bpmIndex)
    #
    #     return newBpms


import unittest
from tests.test.RSC_PATHS import *

from reamber.algorithms.convert.QuaToOsu import QuaToOsu
from reamber.quaver.QuaMapObj import QuaMapObj
# import logging
#
# logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestQuaToOsu(unittest.TestCase):

    # @profile
    def test_qua1(self):
        # Complex BPM Points
        qua = QuaMapObj()
        qua.readFile(QUA_NEURO_CLOUD)

        osu = QuaToOsu.convert(qua)
        osu.writeFile("out.osu")


if __name__ == '__main__':
    unittest.main()



import cProfile

def profile(fun):
    def wrapper(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()
        fun(*args, **kwargs)
        pr.disable()
        # after your program ends
        pr.print_stats(sort="tottime")
    return wrapper



import unittest
from tests.test.RSC_PATHS import *

from reamber.osu.OsuMapObj import OsuMapObj
from reamber.algorithms.meta.hitSoundCopy import hitSoundCopy

import logging

logging.basicConfig(filename="event.log", filemode="w+", level=logging.DEBUG)


class TestHitsoundCopy(unittest.TestCase):

    # # @profile
    # def test_osu1(self):
    #     mFrom = OsuMapObj()
    #     mFrom.readFile(OSU_AVENGER_HITSOUNDFILE)
    #
    #     mTo = OsuMapObj()
    #     mTo.readFile(OSU_AVENGER_HITSOUNDABLE)
    #
    #     mOut = hitSoundCopy(mFrom=mFrom, mTo=mTo)
    #
    #     mOut.writeFile("out.osu")

    # @profile
    def test_osu2(self):
        mFrom = OsuMapObj()
        mFrom.readFile(OSU_TRIBAL_TRIAL_MX)

        mTo = OsuMapObj()
        mTo.readFile(OSU_TRIBAL_TRIAL_EXH)

        mOut = hitSoundCopy(mFrom=mFrom, mTo=mTo)

        mOut.writeFile("out.osu")

        mOut = hitSoundCopy(mFrom=mOut, mTo=mFrom)

        mOut.writeFile("out.osu")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import List
from reamber.base.HitObj import HitObj
from reamber.base.lists.notes.NoteList import NoteList
from abc import ABC


class HitList(List[HitObj], NoteList, ABC):
    """ Deprecated, don't think this is useful """

    def data(self) -> List[HitObj]:
        return self




from __future__ import annotations
from reamber.algorithms.analysis.playField.parts.PFDrawable import PFDrawable
from reamber.algorithms.analysis.bpm.bpmBeatOffsets import bpmBeatOffsets
from reamber.algorithms.analysis.playField import PlayField

from reamber.base.RAConst import RAConst

from typing import List, Dict


class PFDrawBeatLines(PFDrawable):

    def __init__(self,
                 divisions: List = None,
                 defaultColor: str = "#666666",
                 divisionColors: Dict = None):
        """ Draws beat lines by division specified

        Supported Default Snap Colors: 1, 2, 3, 4, 5, 6, 8, 12, 16, 24, 32.

        The colors can be found in RAConst.

        You can specify non-default snaps (including floats) and your custom divisionColors.

        The new colors will override the default colors if they overlap.

        :param divisions: Defaults to [1, 2, 4], will draw 1/1, 1/2, 1/4 lines in the field
        :param defaultColor: Default color to use when the snap color is not supported.
        :param divisionColors: A custom color dictionary to use. This can be specified if you want to override colors. \
        A template can be found in RAConst.DIVISION_COLOR.
        """
        self.divisions = divisions if divisions else [1, 2, 4]  # Default divisions
        self.defaultColor = defaultColor
        # The unpacking operator resolves the dictionary as specified in the docstring
        self.divisionColors = \
            {**RAConst.DIVISION_COLORS, **divisionColors} if divisionColors else RAConst.DIVISION_COLORS

    def draw(self, pf: PlayField) -> PlayField:
        """ Refer to __init__ """
        # Need to draw it from most common to least common, else it'll overlap incorrectly
        for division in sorted(self.divisions, reverse=True):

            if division not in self.divisionColors.keys():
                color = self.defaultColor  # Default color if val not found
            else:
                color = self.divisionColors[division]

            for beat in bpmBeatOffsets(pf.m.bpms, nths=division, lastOffset=pf.m.notes.lastOffset()):
                pf.canvasDraw.line(
                    [(0,                       pf.canvasH - int((beat - pf.start) / pf.durationPerPx)),
                     (pf.canvasW - pf.padding, pf.canvasH - int((beat - pf.start) / pf.durationPerPx))],
                    fill=color)

        return pf



from reamber.sm.SMMapSetObj import SMMapSetObj, SMMapObj
from reamber.o2jam.O2JMapSetObj import O2JMapSetObj
from reamber.base.BpmObj import BpmObj
from reamber.sm.SMMapObjMeta import SMMapObjChartTypes
from reamber.sm.SMHitObj import SMHitObj
from reamber.sm.SMHoldObj import SMHoldObj
from reamber.sm.SMBpmObj import SMBpmObj
from reamber.sm.lists.SMNotePkg import SMNotePkg
from reamber.sm.lists.SMBpmList import SMBpmList
from reamber.sm.lists.notes.SMHitList import SMHitList
from reamber.sm.lists.notes.SMHoldList import SMHoldList
from typing import List


class O2JToSM:
    @staticmethod
    def convert(o2j: O2JMapSetObj) -> List[SMMapSetObj]:
        """ Converts a Mapset to multiple SM maps

        Due to non-confidence that bpms are consistent, A list of SMSet would be generated.

        :param o2j: O2Jam Mapset
        :return: List of SM MapSets
        """

        smSets = []

        for o2jMap in o2j.maps:

            bpms: List[BpmObj] = []
            for bpm in o2jMap.bpms:
                bpms.append(SMBpmObj(offset=bpm.offset, bpm=bpm.bpm))

            hits: List[SMHitObj] = []
            holds: List[SMHoldObj] = []

            for hit in o2jMap.notes.hits():
                hits.append(SMHitObj(offset=hit.offset, column=hit.column))
            for hold in o2jMap.notes.holds():
                holds.append(SMHoldObj(offset=hold.offset, column=hold.column, length=hold.length))

            smSet: SMMapSetObj = SMMapSetObj(
                title=o2j.title,
                artist=o2j.artist,
                credit=o2j.creator,
                offset=0.0,
                maps=[
                    SMMapObj(
                        description=f"Level {o2j.level[o2j.maps.index(o2jMap)]}",
                        chartType=SMMapObjChartTypes.KB7_SINGLE,
                        notes=SMNotePkg(hits=SMHitList(hits),
                                        holds=SMHoldList(holds)),
                        bpms=SMBpmList(bpms)
                    )
                ]
            )
            smSets.append(smSet)

        return smSets



from __future__ import annotations
from reamber.base.TimedObj import TimedObj
from dataclasses import dataclass


@dataclass
class OsuSampleObj(TimedObj):
    """ Osu Samples are automatically played hitsounds. Under [Events]

    Not to be confused with OsuSampleSet, where that's a class of static variables
    """

    sampleFile: str = ""
    volume: int = 70  # Osu defaults all samples to 70

    @staticmethod
    def readString(s: str) -> OsuSampleObj or None:
        """ Reads the string as a sample """

        if s.isspace():
            return None

        sComma = s.split(",")
        if len(sComma) < 5:
            return None

        this = OsuSampleObj()
        this.offset = float(sComma[1])
        this.sampleFile = sComma[3]
        this.volume = int(sComma[4])

        return this

    def writeString(self) -> str:
        """ Exports the sample as a string

        e.g. Sample,1600,0,"01.wav",70
        """
        return f"Sample,{int(self.offset)},0,{self.sampleFile},{self.volume}"



from reamber.algorithms.analysis.describe.describe import describe, describeNotes
from reamber.algorithms.analysis.describe.meta import mapMetadata

__all__ = ['describe', 'describeNotes', 'mapMetadata']
